"""
A module for computing Halo Occupation Distribution (HOD) models.
This module provides classes and functions to calculate properties of galaxies within dark matter halos,
using various HOD models and conditional observable functions (COFs).
"""

import numpy as np
from scipy.integrate import simpson
from scipy.interpolate import interp1d
from scipy.special import erf

from hmf._internals._framework import Component, pluggable

valid_units = ['1/h', '1/h^2']
hod_settings_defaults = {
    'observables_file': None,
    'obs_min': np.atleast_1d(8.0),
    'obs_max': np.atleast_1d(12.0),
    'zmin': np.atleast_1d(0.0),
    'zmax': np.atleast_1d(0.2),
    'nz': 15,
    'nobs': 300,
    'observable_h_unit': '1/h^2',
}


def load_data(file_name):
    """
    Load data from a file.

    Parameters:
    -----------
    file_name : str
        Path to the file to load data from.

    Returns:
    --------
    tuple
        Redshift data, minimum observable data, and maximum observable data.
    """
    z_data, obs_min, obs_max = np.loadtxt(
        file_name, usecols=(0, 1, 2), unpack=True, dtype=float
    )
    return z_data, obs_min, obs_max


@pluggable
class HaloOccupationDistribution(Component):
    r"""
    Base class for Halo Occupation Distribution (HOD) models.
    This class provides the framework for computing various properties of galaxies within dark matter halos.

    Parameters:
    -----------
    cosmo : object
        Astropy model object.
    mass : array_like
        Array of halo masses.
    dndlnm : array_like
        Halo mass function.
    halo_bias : array_like
        Halo bias.
    z_vec : array_like
        Array of redshifts.
    hod_settings : dict
        Dictionary of HOD settings.
    \*\*model_parameters :
        These parameters are model-specific. For any model, list the available
        parameters (and their defaults) using ``<model>._defaults``
    """

    _defaults = {}

    def __init__(
        self,
        cosmo=None,
        mass=None,
        dndlnm=None,
        halo_bias=None,
        z_vec=None,
        hod_settings=None,
        **model_parameters,
    ):
        self.cosmo = cosmo
        self.mass = mass[np.newaxis, np.newaxis, :]
        self.z_vec = z_vec
        self.hod_settings = hod_settings or hod_settings_defaults
        self.dndlnm = dndlnm
        self.halo_bias = halo_bias

        super().__init__(**model_parameters)

    @property
    def obs(self):
        """
        Returns the array of observables at which to calculate the CSMF/CLF and integrate over to get the HOD

        Returns:
        --------
        array_like
            array of observables
        """
        obs = np.array(
            [
                [
                    np.logspace(
                        self.log_obs_min[nb, jz], self.log_obs_max[nb, jz], self.nobs
                    )
                    for jz in range(self.nz)
                ]
                for nb in range(self.nbins)
            ]
        )
        # With newaxis we make sure the COF shape is (nb, nz, nmass, nobs)
        return obs[:, :, np.newaxis, :]

    @property
    def dndlnm_int(self):
        """
        Returns the interpolated halo mass function at HOD specific redshifts.

        Returns:
        --------
        array_like
            interpolated halo mass function at self.z
        """
        dndlnm_fnc = interp1d(
            self.z_vec,
            self.dndlnm,
            kind='linear',
            fill_value='extrapolate',
            bounds_error=False,
            axis=0,
        )
        return dndlnm_fnc(self.z)

    @property
    def halo_bias_int(self):
        """
        Returns the interpolated halo bias function at HOD specific redshifts.

        Returns:
        --------
        array_like
            interpolated halo bias function at self.z
        """
        halo_bias_fnc = interp1d(
            self.z_vec,
            self.halo_bias,
            kind='linear',
            fill_value='extrapolate',
            bounds_error=False,
            axis=0,
        )
        return halo_bias_fnc(self.z)

    @property
    def data(self):
        """
        Returns the z_bins, obs_min and obs_max quantities from tabulated observables file.

        Returns:
        --------
        tupple or None
        """
        if self.hod_settings['observables_file'] is not None:
            z_bins, obs_min, obs_max = load_data(self.hod_settings['observables_file'])
            return z_bins, obs_min, obs_max
        else:
            return None

    @property
    def nobs(self):
        """
        Sets the number of observables in the observable array.

        Returns:
        --------
        array_like
            number of obserables in the observable array
        """
        return self.hod_settings['nobs']

    @property
    def nbins(self):
        """
        Returns the number of HOD bins.

        Returns:
        --------
        int
            number of HOD bins
        """
        if self.hod_settings['observables_file'] is not None:
            return 1
        else:
            return len(self.hod_settings['obs_min'])

    @property
    def nz(self):
        """
        Sets the number of redshift bins.

        Returns:
        --------
        int
            number of redshift bins
        """
        if self.hod_settings['observables_file'] is not None:
            return len(self.data[0])
        else:
            return self.hod_settings['nz']

    @property
    def z(self):
        """
        Sets the HOD specific redshift array.

        Returns:
        --------
        array_like
            HOD specific redshifts
        """
        if self.hod_settings['observables_file'] is not None:
            return self.data[0][np.newaxis, :]
        else:
            zmin = self.hod_settings['zmin']
            zmax = self.hod_settings['zmax']
            return np.array(
                [
                    np.linspace(zmin_i, zmax_i, self.nz)
                    for zmin_i, zmax_i in zip(zmin, zmax, strict=False)
                ]
            )

    @property
    def log_obs_min(self):
        """
        Sets the min observable limits.

        Returns:
        --------
        array_like
            min observable limits
        """
        if self.hod_settings['observables_file'] is not None:
            return np.log10(self.data[1])[np.newaxis, :]
        else:
            obs_min = self.hod_settings['obs_min']
            return np.array([np.repeat(obs_min_i, self.nz) for obs_min_i in obs_min])

    @property
    def log_obs_max(self):
        """
        Sets the max observable limits.

        Returns:
        --------
        array_like
            max observable limits
        """
        if self.hod_settings['observables_file'] is not None:
            return np.log10(self.data[2])[np.newaxis, :]
        else:
            obs_max = self.hod_settings['obs_max']
            return np.array([np.repeat(obs_max_i, self.nz) for obs_max_i in obs_max])

    def _mass_integral(self, hod):  # pragma: no cover
        """
        Compute the mass integral for a given HOD.

        Parameters:
        -----------
        hod : array_like
            Halo Occupation Distribution.

        Returns:
        --------
        array_like
            Integral of the HOD weighted by the halo mass function.
        """
        integrand = hod * self.dndlnm_int / self.mass
        return simpson(integrand, x=self.mass, axis=-1)

    def _mean_mass_integral(self, hod):  # pragma: no cover
        """
        Compute the mean mass integral for a given HOD.

        Parameters:
        -----------
        hod : array_like
            Halo Occupation Distribution.

        Returns:
        --------
        array_like
            Integral of the HOD weighted by the halo mass function and halo mass.
        """
        integrand = hod * self.dndlnm_int
        return simpson(integrand, x=self.mass, axis=-1)

    def _bias_integral(self, hod):  # pragma: no cover
        """
        Compute the bias integral for a given HOD.

        Parameters:
        -----------
        hod : array_like
            Halo Occupation Distribution.

        Returns:
        --------
        array_like
            Integral of the HOD weighted by the halo bias and halo mass function.
        """
        bg_integrand = hod * self.halo_bias_int * self.dndlnm_int / self.mass
        return simpson(bg_integrand, x=self.mass, axis=-1) / self._ntot

    def _interpolate(self, data, fill_value='extrapolate', axis=-1):
        """
        Helper function to interpolate data along a given axis in an array.

        Parameters:
        -----------
        data : array_like
            Data to interpolate.
        fill_value : str, optional
            Value to use for points outside the data range.
        axis : int, optional
            Axis along which to interpolate.

        Returns:
        --------
        array_like
            Interpolated data.
        """
        n_int = [
            interp1d(
                self.z[i], data[i], fill_value=fill_value, bounds_error=False, axis=axis
            )
            for i in range(self.z.shape[0])
        ]
        return np.array([f(self.z_vec) for f in n_int])

    @property
    def _ncen(self):
        r"""
        Total number density of central galaxies with the given HOD.
        This is an integral over the HOD and the halo mass function to remove the halo mass dependence.

        :math:`N_{\rm x} = \int ⟨N_{rm x}|M⟩ n(M) {\rm d}M`
        """
        return self._mass_integral(self._compute_hod_cen)

    @property
    def _nsat(self):
        r"""
        Total number density of satellite galaxies with the given HOD.
        This is an integral over the HOD and the halo mass function to remove the halo mass dependence.

        :math:`N_{\rm x} = \int ⟨N_{rm x}|M⟩ n(M) {\rm d}M`
        """
        return self._mass_integral(self._compute_hod_sat)

    @property
    def _ntot(self):
        r"""
        Total number density of galaxies with the given HOD.
        This is an integral over the HOD and the halo mass function to remove the halo mass dependence.

        :math:`N_{\rm x} = \int ⟨N_{rm x}|M⟩ n(M) {\rm d}M`
        """
        return self._mass_integral(self._compute_hod)

    @property
    def _mass_avg_cen(self):
        r"""
        The mean halo mass for the given population of galaxies, e.g. central and satellite galaxies.

        :math:`M_{\rm mean} = \int ⟨N_{_rm x}|M⟩ M n(M) {\rm d}M`
        """
        return self._mean_mass_integral(self._compute_hod_cen)

    @property
    def _mass_avg_sat(self):
        r"""
        The mean halo mass for the given population of galaxies, e.g. central and satellite galaxies.

        :math:`M_{\rm mean} = \int ⟨N_{_rm x}|M⟩ M n(M) {\rm d}M`
        """
        return self._mean_mass_integral(self._compute_hod_sat)

    @property
    def _mass_avg_tot(self):
        r"""
        The mean halo mass for the given population of galaxies, e.g. central and satellite galaxies.

        :math:`M_{\rm mean} = \int ⟨N_{_rm x}|M⟩ M n(M) {\rm d}M`
        """
        return self._mean_mass_integral(self._compute_hod)

    @property
    def _bg_cen(self):
        r"""
        Mean linear halo bias for the given population of galaxies.

        :math:`b_{\rm lin, x} = \int ⟨N_{\rm x}|M⟩ b_h(M) n(M) {\rm d}M`
        """
        return self._bias_integral(self._compute_hod_cen)

    @property
    def _bg_sat(self):
        r"""
        Mean linear halo bias for the given population of galaxies.

        :math:`b_{\rm lin, x} = \int ⟨N_{\rm x}|M⟩ b_h(M) n(M) {\rm d}M`
        """
        return self._bias_integral(self._compute_hod_sat)

    @property
    def _bg_tot(self):
        r"""
        Mean linear halo bias for the given population of galaxies.

        :math:`b_{\rm lin, x} = \int ⟨N_{\rm x}|M⟩ b_h(M) n(M) {\rm d}M`
        """
        return self._bias_integral(self._compute_hod)

    @property
    def number_density_cen(self):
        """
        Compute the number density of central galaxies.

        Returns:
        --------
        array_like
            number density of central galaxies
        """
        return self._interpolate(self._ncen)

    @property
    def number_density_sat(self):
        """
        Compute the number density of satellite galaxies.

        Returns:
        --------
        array_like
            number density of satellite galaxies
        """
        return self._interpolate(self._nsat)

    @property
    def number_density(self):
        """
        Compute the number density of galaxies.

        Returns:
        --------
        array_like
            number density of all galaxies
        """
        return self._interpolate(self._ntot)

    @property
    def f_c(self):
        """
        Fraction of central galaxies.

        Returns:
        --------
        array_like
            fraction of central galaxies
        """
        f_c = self._ncen / self._ntot
        return self._interpolate(f_c, fill_value=0.0)

    @property
    def f_s(self):
        """
        Fraction of satellite galaxies.

        Returns:
        --------
        array_like
            fraction of satellite galaxies
        """
        f_s = self._nsat / self._ntot
        return self._interpolate(f_s, fill_value=0.0)

    @property
    def avg_halo_mass_cen(self):
        """
        Compute the average halo mass for central galaxies.

        Returns:
        --------
        array_like
            average halo mass for central galaxies
        """
        return self._interpolate(self._mass_avg_cen, fill_value=0.0)

    @property
    def avg_halo_mass_sat(self):
        """
        Compute the average halo mass for satellite galaxies.

        Returns:
        --------
        array_like
            average halo mass for satellite galaxies
        """
        return self._interpolate(self._mass_avg_sat, fill_value=0.0)

    @property
    def avg_halo_mass(self):
        """
        Compute the average halo mass for galaxies.

        Returns:
        --------
        array_like
            average halo mass for all galaxies
        """
        return self._interpolate(self._mass_avg_tot, fill_value=0.0)

    @property
    def galaxy_linear_bias_cen(self):
        """
        Compute the galaxy linear bias for central galaxies.

        Returns:
        --------
        array_like
            galaxy linear bias for central galaxies
        """
        return self._interpolate(self._bg_cen)

    @property
    def galaxy_linear_bias_sat(self):
        """
        Compute the galaxy linear bias for satellite galaxies.

        Returns:
        --------
        array_like
            galaxy linear bias for satellite galaxies
        """
        return self._interpolate(self._bg_sat)

    @property
    def galaxy_linear_bias(self):
        """
        Compute the galaxy linear bias for galaxies.

        Returns:
        --------
        array_like
            galaxy linear bias for all galaxies
        """
        return self._interpolate(self._bg_tot)

    @property
    def hod_cen(self):
        """
        Compute the HOD for central galaxies.

        Returns:
        --------
        array_like
            HOD of central galaxies
        """
        return self._interpolate(self._compute_hod_cen, axis=0)

    @property
    def hod_sat(self):
        """
        Compute the HOD for satellite galaxies.

        Returns:
        --------
        array_like
            HOD of satellite galaxies
        """
        return self._interpolate(self._compute_hod_sat, axis=0)

    @property
    def hod(self):
        """
        Compute the HOD for galaxies.

        Returns:
        --------
        array_like
            HOD of all galaxies
        """
        return self._interpolate(self._compute_hod, axis=0)

    @property
    def stellar_fraction_cen(self):
        """
        Compute the stellar fraction for central galaxies.

        Returns:
        --------
        array_like
            stellar fraction of central galaxies
        """
        if self._compute_stellar_fraction_cen is None:
            return np.zeros((self.nbins, self.z_vec.size, self.mass.shape[-1]))
        else:
            fstar = self._interpolate(self._compute_stellar_fraction_cen, axis=0)
            if self.hod_settings['observable_h_unit'] == valid_units[1]:
                fstar = fstar * self.cosmo.h
            return fstar

    @property
    def stellar_fraction_sat(self):
        """
        Compute the stellar fraction for satellite galaxies.

        Returns:
        --------
        array_like
            stellar fraction of satellite galaxies
        """
        if self._compute_stellar_fraction_sat is None:
            return np.zeros((self.nbins, self.z_vec.size, self.mass.shape[-1]))
        else:
            fstar = self._interpolate(self._compute_stellar_fraction_sat, axis=0)
            if self.hod_settings['observable_h_unit'] == valid_units[1]:
                fstar = fstar * self.cosmo.h
            return fstar

    @property
    def stellar_fraction(self):
        """
        Compute the stellar fraction for galaxies.

        Returns:
        --------
        array_like
            stellar fraction of all galaxies
        """
        if self._compute_stellar_fraction is None:
            return np.zeros((self.nbins, self.z_vec.size, self.mass.shape[-1]))
        else:
            fstar = self._interpolate(self._compute_stellar_fraction, axis=0)
            if self.hod_settings['observable_h_unit'] == valid_units[1]:
                fstar = fstar * self.cosmo.h
            return fstar

    @property
    def _compute_hod_cen(self):
        """
        Compute the HOD for central galaxies.
        """
        raise NotImplementedError

    @property
    def _compute_hod_sat(self):
        """
        Compute the HOD for satellite galaxies.
        """
        raise NotImplementedError

    @property
    def _compute_hod(self):
        """
        Compute the total HOD by summing central and satellite HODs.
        """
        return self._compute_hod_cen + self._compute_hod_sat

    @property
    def _compute_stellar_fraction_cen(self):
        return None

    @property
    def _compute_stellar_fraction_sat(self):
        return None

    @property
    def _compute_stellar_fraction(self):
        return None


class Cacciato(HaloOccupationDistribution):
    r"""
    CSMF/CLF model from Cacciato et al. (2013) [1]_.

    The conditional observable functions (COFs) tell us how many galaxies with the observed property O,
    exist in haloes of mass M: :math:`\Phi(O|M)`.

    Integrating over the observable will give us the total number of galaxies in haloes of a given mass,
    the so-called Halo Occupation Distribution (HOD).

    The observable can be galaxy stellar mass or galaxy luminosity or possibly other properties of galaxies.
    Note that the general mathematical form of the COFs might not hold for other observables.

    COF is different for central and satellite galaxies. The total COF can be written as the sum of the two:

    :math:`\Phi(O|M) = \Phi_{\rm c}(O|M) + \Phi_{\rm s}(O|M)`

    The halo mass dependence comes in through pivot observable values denoted by :math:`\star`, e.g. :math:`O_{\star, {\rm c}}`, :math:`O_{\star, {\rm s}}`

    Parameters:
    -----------
    log10_obs_norm_c : float
        Log10 of the normalization for central galaxies.
    log10_m_ch : float
        Log10 of the characteristic mass.
    g1 : float
        Low mass slope parameter for central galaxies.
    g2 : float
        High mass slope parameter for central galaxies.
    sigma_log10_O_c : float
        Scatter in log10 of the observable for central galaxies.
    norm_s : float
        Normalization for satellite galaxies.
    pivot : float
        Pivot mass for the normalization of the stellar mass function.
    alpha_s : float
        Slope parameter for satellite galaxies.
    beta_s : float
        Exponent parameter for satellite galaxies.
    b0 : float
        Parameter for the conditional stellar mass function.
    b1 : float
        Parameter for the conditional stellar mass function.
    b2 : float
        Parameter for the conditional stellar mass function.
    A_cen : float, optional
        Decorated HOD assembly bias parameter for central galaxies.
    A_sat : float, optional
        Decorated HOD assembly bias parameter for satellite galaxies.

    References
    ----------
    .. [1]  Cacciato, M. et al., "Cosmological constraints from a combination of galaxy clustering and lensing - III. Application to SDSS data",
            https://academic.oup.com/mnras/article/430/2/767/2891826.
    """

    # all observable masses in units of log10(M_sun h^-2)
    _defaults = {
        'log10_obs_norm_c': 9.95,  # O_0, O_norm_c
        'log10_m_ch': 11.24,  # log10_m_ch
        'g1': 3.18,  # gamma_1
        'g2': 0.245,  # gamma_2
        'sigma_log10_O_c': 0.157,  # sigma_log10_O_c
        'norm_s': 0.562,  # extra normalisation factor for satellites
        'pivot': 12.0,  # pivot mass for the normalisation of the stellar mass function: ϕ∗s
        'alpha_s': -1.18,  # goes into the conditional stellar mass function COF_sat(M*|M)
        'beta_s': 2,  # goes into the conditional stellar mass function COF_sat(M*|M)
        'b0': -1.17,  # log10[ϕ∗s(M)] = b0 + b1(log10 m_p)+ b2(log10 m_p)^2, m_p = M/pivot
        'b1': 1.53,  # log10[ϕ∗s(M)] = b0 + b1(log10 m_p)+ b2(log10 m_p)^2, m_p = M/pivot
        'b2': -0.217,  # log10[ϕ∗s(M)] = b0 + b1(log10 m_p)+ b2(log10 m_p)^2, m_p = M/pivot
        'A_cen': None,  # Decorated HOD assembly bias parameters
        'A_sat': None,  # Decorated HOD assembly bias parameters
    }

    @property
    def M_char(self):
        """
        Return the 10.**log10_m_ch

        Returns:
        --------
        array_like
            10.**log10_m_ch
        """
        return 10.0 ** self.params['log10_m_ch']

    @property
    def Obs_norm_c(self):
        """
        Return the 10.**log10_obs_norm_c

        Returns:
        --------
        array_like
            10.**log10_obs_norm_c
        """
        return 10.0 ** self.params['log10_obs_norm_c']

    @property
    def COF_cen(self):
        r"""
        COF for Central galaxies.

        :math:`\Phi_{\rm c}(O|M) = 1/[\sqrt{(2\pi)} \ln(10) \sigma_{\rm c} O] \exp[-\log(O/O_{\star, {\rm c}})^{2}/ (2 \sigma_{\rm c}^{2})]`

        Note  :math:`\Phi_{\rm c}(O|M)` is unitless.

        Returns:
        --------
        array_like
            COF of central galaxies
        """
        mean_obs_c = self.cal_mean_obs_c[:, :, :, np.newaxis]  # O∗c
        COF_c = (
            1.0
            / (
                np.sqrt(2.0 * np.pi)
                * np.log(10.0)
                * self.params['sigma_log10_O_c']
                * self.obs
            )
            * np.exp(
                -((np.log10(self.obs / mean_obs_c)) ** 2)
                / (2.0 * self.params['sigma_log10_O_c'] ** 2)
            )
        )
        return COF_c

    @property
    def COF_sat(self):
        r"""
        COF for satellite galaxies.

        :math:`\Phi_{\rm s}(O|M) = \phi_{\star, {\rm s}}/O_{\star, {\rm s}} (O/O_{\star, {\rm s}})^{\alpha_{\rm s}} \exp [-(O/O_{\star, {\rm s}})^{2}]`,

        :math:`O_{\star, {\rm s}}` is :math:`O_{\star, {\rm s}}(M) = 0.56 O_{\star, {\rm c}}(M)`
        Note :math:`\Phi_{\rm s}(O|M)` is unitless.

        Returns:
        --------
        array_like
            COF of satellite galaxies
        """
        obs_s_star = self.params['norm_s'] * self.cal_mean_obs_c[:, :, :, np.newaxis]
        obs_tilde = self.obs / obs_s_star
        phi_star_val = self.phi_star_s[:, :, :, np.newaxis]
        COF_s = (
            (phi_star_val / obs_s_star)
            * (obs_tilde ** self.params['alpha_s'])
            * np.exp(-(obs_tilde ** self.params['beta_s']))
        )
        return COF_s

    @property
    def COF(self):
        """
        Total COF, sum of central and satellite COFs.

        Returns:
        --------
        array_like
            COF of all galaxies
        """
        return self.COF_cen + self.COF_sat

    @property
    def obs_func_cen(self):
        r"""
        The observable function (SMF/LF) for central galaxies.
        Defined as: :math:`\Phi_{\rm c}(O|M)`, :math:`\Phi_{\rm s}(O|M)` integrated over the halo mass weighted
        with the Halo Mass Function (HMF) to give:  :math:`\Phi_{\rm c}(O)`, :math:`\Phi_{\rm c}(O)`

        :math:`\Phi_{\rm x}(O) = \int \Phi_{\rm x}(O|M) n(M) {\rm d}M`,

        dndlnm is basically n(M) x mass, it is the output of hmf
        The differential mass function in terms of natural log of m,
        len=len(m) [units \(h^3 Mpc^{-3}\)]

        obs_func unit is h^3 Mpc^{-3} dex^-1

        Returns:
        --------
        array_like
            observable function of central galaxies
        """
        integrand = (
            self.COF_cen
            * self.dndlnm_int[:, :, :, np.newaxis]
            / self.mass[:, :, :, np.newaxis]
        )
        obs_function = simpson(integrand, x=self.mass[:, :, :, np.newaxis], axis=-2)
        return obs_function

    @property
    def obs_func_sat(self):
        r"""
        The observable function (SMF/LF) for satellite galaxies.
        Defined as: :math:`\Phi_{\rm c}(O|M)`, :math:`\Phi_{\rm s}(O|M)` integrated over the halo mass weighted
        with the Halo Mass Function (HMF) to give:  :math:`\Phi_{\rm c}(O)`, :math:`\Phi_{\rm c}(O)`

        :math:`\Phi_{\rm x}(O) = \int \Phi_{\rm x}(O|M) n(M) {\rm d}M`,

        dndlnm is basically n(M) x mass, it is the output of hmf
        The differential mass function in terms of natural log of m,
        len=len(m) [units \(h^3 Mpc^{-3}\)]

        obs_func unit is h^3 Mpc^{-3} dex^-1

        Returns:
        --------
        array_like
            observable function of satellite galaxies
        """
        integrand = (
            self.COF_sat
            * self.dndlnm_int[:, :, :, np.newaxis]
            / self.mass[:, :, :, np.newaxis]
        )
        obs_function = simpson(integrand, x=self.mass[:, :, :, np.newaxis], axis=-2)
        return obs_function

    @property
    def obs_func(self):
        r"""
        The observable function (SMF/LF).
        Defined as: :math:`\Phi_{\rm c}(O|M)`, :math:`\Phi_{\rm s}(O|M)` integrated over the halo mass weighted
        with the Halo Mass Function (HMF) to give:  :math:`\Phi_{\rm c}(O)`, :math:`\Phi_{\rm c}(O)`

        :math:`\Phi_{\rm x}(O) = \int \Phi_{\rm x}(O|M) n(M) {\rm d}M`,

        dndlnm is basically n(M) x mass, it is the output of hmf
        The differential mass function in terms of natural log of m,
        len=len(m) [units \(h^3 Mpc^{-3}\)]

        obs_func unit is h^3 Mpc^{-3} dex^-1

        Returns:
        --------
        array_like
            observable function of all galaxies
        """
        integrand = (
            self.COF
            * self.dndlnm_int[:, :, :, np.newaxis]
            / self.mass[:, :, :, np.newaxis]
        )
        obs_function = simpson(integrand, x=self.mass[:, :, :, np.newaxis], axis=-2)
        return obs_function

    @property
    def cal_mean_obs_c(self):
        r"""
        Stellar to halo mass relation (observable to halo mass relation).

        :math:`O_{\star, {\rm c}}(M) = O_0 (M/M_{1})^{\gamma_{1}} / [1 + (M/M_{1})]^{(\gamma_{1} - \gamma_{2})}`

        To get the values for the satellite call this * hod_par.norm_s

        :math:`O_{\star, {\rm s}}(M) = 0.56 O_{\star, {\rm c}}(M)`

        Here  :math:`M_1` is a characteristic mass scale, and :math:`O_0` is the normalization.

        (observable can be galaxy luminosity or stellar mass)
        returns the observable given halo mass.
        Eqs 19 of D23: 2210.03110

        Returns:
        --------
        array_like
            observable given halo mass - SMHM relation
        """
        mean_obs_c = (
            self.Obs_norm_c
            * (self.mass / self.M_char) ** self.params['g1']
            / (1.0 + (self.mass / self.M_char))
            ** (self.params['g1'] - self.params['g2'])
        )
        return mean_obs_c

    @property
    def phi_star_s(self):
        r"""
        Normalisation of satellite COF function.

        :math:`log[\phi_{\star, {\rm s}}(M)] = b_0 + b_1(\log M_{13})` , :math:`M_{13} = M/({\rm pivot})`

        Eqs 21 and 22 of D23: 2210.03110

        Returns:
        --------
        array_like
            normalisation of satellite COF function
        """
        logM_pivot = np.log10(self.mass) - self.params['pivot']
        log_phi_s = (
            self.params['b0']
            + self.params['b1'] * logM_pivot
            + self.params['b2'] * (logM_pivot**2.0)
        )
        return 10.0**log_phi_s

    @property
    def _compute_hod_cen(self):
        r"""
        Eq 23 of D23: 2210.03110

        :math:`\langle N_{\rm x}|M \rangle = \int_{O_{\rm low}}^{O_{\rm high}} \Phi_{\rm x}(O|M) {\rm d}O`
        """
        N_cen = simpson(self.COF_cen, x=self.obs)
        if self.params['A_cen'] is not None:
            delta_pop_c = self.params['A_cen'] * np.fmin(N_cen, 1.0 - N_cen)
            N_cen = N_cen + delta_pop_c
        return N_cen

    @property
    def _compute_hod_sat(self):
        r"""
        Eq 23 of D23: 2210.03110

        :math:`\langle N_{\rm x}|M \rangle = \int_{O_{\rm low}}^{O_{\rm high}} \Phi_{\rm x}(O|M) {\rm d}O`
        """
        N_sat = simpson(self.COF_sat, x=self.obs)
        if self.params['A_sat'] is not None:
            delta_pop_s = self.params['A_sat'] * N_sat
            N_sat = N_sat + delta_pop_s
        return N_sat

    @property
    def _compute_stellar_fraction_cen(self):
        r"""
        The mean value of the observable for the given galaxy population for a given halo mass.
        O is weighted by the number of galaxies with the property O for each halo mass: :math:`\Phi_{\rm x}(O|M)`

        :math:`f_{\star} = \int_{O_{\rm low}}^{O_{\rm high}} \Phi_{\rm x}(O|M) O {\rm d}O`
        """
        return simpson(self.COF_cen * self.obs, x=self.obs) / self.mass

    @property
    def _compute_stellar_fraction_sat(self):
        r"""
        The mean value of the observable for the given galaxy population for a given halo mass.
        O is weighted by the number of galaxies with the property O for each halo mass: :math:`\Phi_{\rm x}(O|M)`

        :math:`f_{\star} = \int_{O_{\rm low}}^{O_{\rm high}} \Phi_{\rm x}(O|M) O {\rm d}O`
        """
        return simpson(self.COF_sat * self.obs, x=self.obs) / self.mass

    @property
    def _compute_stellar_fraction(self):
        r"""
        The mean value of the observable for the given galaxy population for a given halo mass.
        O is weighted by the number of galaxies with the property O for each halo mass: :math:`\Phi_{\rm x}(O|M)`

        :math:`f_{\star} = \int_{O_{\rm low}}^{O_{\rm high}} \Phi_{\rm x}(O|M) O {\rm d}O`
        """
        return simpson(self.COF * self.obs, x=self.obs) / self.mass


class Simple(HaloOccupationDistribution):
    """
    Simple HOD model

    Parameters:
    -----------
    log10_Mmin : array_like
        Log10 of the minimum mass for central galaxies.
    log10_Msat : array_like
        Log10 of the minimum mass for satellite galaxies.
    alpha : array_like
        Slope parameter for satellite galaxies.
    A_cen : float, optional
        Decorated HOD assembly bias parameter for central galaxies.
    A_sat : float, optional
        Decorated HOD assembly bias parameter for satellite galaxies.
    """

    _defaults = {
        'log10_Mmin': 12.0,
        'log10_Msat': 13.0,
        'alpha': 1.0,
        'A_cen': None,
        'A_sat': None,
    }

    @property
    def log10_Mmin(self):
        """
        Log10 of the minimum mass for central galaxies.

        :type: array_like
        """
        val = self.params['log10_Mmin']
        if not hasattr(val, '__len__'):
            val = [val]
        return np.array(val)

    @property
    def log10_Msat(self):
        """
        Log10 of the minimum mass for satellite galaxies.

        :type: array_like
        """
        val = self.params['log10_Msat']
        if not hasattr(val, '__len__'):
            val = [val]
        return np.array(val)

    @property
    def alpha(self):
        """
        Slope parameter for satellite galaxies.

        :type: array_like
        """
        val = self.params['alpha']
        if not hasattr(val, '__len__'):
            val = [val]
        return np.array(val)[:, np.newaxis, np.newaxis]

    @property
    def Mmin(self):
        """
        Return the Mmin parameter in the correct shape and units.

        Returns:
        --------
        array_like
            Mmin
        """
        return 10.0 ** self.log10_Mmin[:, np.newaxis, np.newaxis]

    @property
    def Msat(self):
        """
        Return the Msat parameter in the correct shape and units.

        Returns:
        --------
        array_like
            Msat
        """
        return 10.0 ** self.log10_Msat[:, np.newaxis, np.newaxis]

    @property
    def _compute_hod_cen(self):
        """
        Compute the HOD for central galaxies.
        """
        N_cen = np.heaviside(self.mass - self.Mmin, 1.0)
        if self.params['A_cen'] is not None:
            delta_pop_c = self.params['A_cen'] * np.fmin(N_cen, 1.0 - N_cen)
            N_cen = N_cen + delta_pop_c
        return np.tile(N_cen, (self.nz, 1))

    @property
    def _compute_hod_sat(self):
        """
        Compute the HOD for satellite galaxies.
        """
        N_sat = self._compute_hod_cen * (self.mass / self.Msat) ** self.alpha
        if self.params['A_sat'] is not None:
            delta_pop_s = self.params['A_sat'] * N_sat
            N_sat = N_sat + delta_pop_s
        return N_sat


class Zehavi(HaloOccupationDistribution):
    """
    HOD model from Zehavi et al. (2004) [1]_.

    Same as Zheng model in the limit that sigma=0 and M0=0
    Mean number of central galaxies is only ever 0 or 1 in this HOD

    Parameters:
    -----------
    log10_Mmin : array_like
        Log10 of the minimum mass for central galaxies.
    log10_Msat : array_like
        Log10 of the minimum mass for satellite galaxies.
    alpha : array_like
        Slope parameter for satellite galaxies.
    A_cen : float, optional
        Decorated HOD assembly bias parameter for central galaxies.
    A_sat : float, optional
        Decorated HOD assembly bias parameter for satellite galaxies.

    References
    ----------
    .. [1]  Zheng, Z. et al., "Galaxy Evolution from Halo Occupation Distribution Modeling of DEEP2 and SDSS Galaxy Clustering",
            https://iopscience.iop.org/article/10.1086/521074.
    """

    _defaults = {
        'log10_Mmin': 12.0,
        'log10_Msat': 13.0,
        'alpha': 1.0,
        'A_cen': None,
        'A_sat': None,
    }

    @property
    def log10_Mmin(self):
        """
        Log10 of the minimum mass for central galaxies.

        :type: array_like
        """
        val = self.params['log10_Mmin']
        if not hasattr(val, '__len__'):
            val = [val]
        return np.array(val)

    @property
    def log10_Msat(self):
        """
        Log10 of the minimum mass for satellite galaxies.

        :type: array_like
        """
        val = self.params['log10_Msat']
        if not hasattr(val, '__len__'):
            val = [val]
        return np.array(val)

    @property
    def alpha(self):
        """
        Slope parameter for satellite galaxies.

        :type: array_like
        """
        val = self.params['alpha']
        if not hasattr(val, '__len__'):
            val = [val]
        return np.array(val)[:, np.newaxis, np.newaxis]

    @property
    def Mmin(self):
        """
        Return the Mmin parameter in the correct shape and units.

        Returns:
        --------
        array_like
            Mmin
        """
        return 10.0 ** self.log10_Mmin[:, np.newaxis, np.newaxis]

    @property
    def Msat(self):
        """
        Return the Msat parameter in the correct shape and units.

        Returns:
        --------
        array_like
            Msat
        """
        return 10.0 ** self.log10_Msat[:, np.newaxis, np.newaxis]

    @property
    def _compute_hod_cen(self):
        """
        Compute the HOD for central galaxies.
        """
        N_cen = np.heaviside(self.mass - self.Mmin, 1.0)
        if self.params['A_cen'] is not None:
            delta_pop_c = self.params['A_cen'] * np.fmin(N_cen, 1.0 - N_cen)
            N_cen = N_cen + delta_pop_c
        return np.tile(N_cen, (self.nz, 1))

    @property
    def _compute_hod_sat(self):
        """
        Compute the HOD for satellite galaxies.
        """
        N_sat = (self.mass / self.Msat) ** self.alpha
        if self.params['A_sat'] is not None:
            delta_pop_s = self.params['A_sat'] * N_sat
            N_sat = N_sat + delta_pop_s
        return np.tile(N_sat, (self.nz, 1))


class Zheng(HaloOccupationDistribution):
    """
    HOD model from Zheng et al. (2005) [1]_.

    Parameters:
    -----------
    log10_Mmin : array_like
        Log10 of the minimum mass for central galaxies.
    log10_M0 : array_like
        Log10 of the cutoff mass for satellite galaxies.
    log10_M1 : array_like
        Log10 of the normalization mass for satellite galaxies.
    sigma : array_like
        Scatter in the central galaxy occupation.
    alpha : array_like
        Slope parameter for satellite galaxies.
    A_cen : float, optional
        Decorated HOD assembly bias parameter for central galaxies.
    A_sat : float, optional
        Decorated HOD assembly bias parameter for satellite galaxies.

    References
    ----------
    .. [1]  Zheng, Z. et al., "Theoretical Models of the Halo Occupation Distribution: Separating Central and Satellite Galaxies",
            https://iopscience.iop.org/article/10.1086/466510.
    """

    _defaults = {
        'log10_Mmin': 12.0,
        'log10_M0': 12.0,
        'log10_M1': 13.0,
        'sigma': 0.15,
        'alpha': 1.0,
        'A_cen': None,
        'A_sat': None,
    }

    @property
    def log10_Mmin(self):
        """
        Log10 of the minimum mass for central galaxies.

        :type: array_like
        """
        val = self.params['log10_Mmin']
        if not hasattr(val, '__len__'):
            val = [val]
        return np.array(val)

    @property
    def log10_M0(self):
        """
        Log10 of the cutoff mass for satellite galaxies.

        :type: array_like
        """
        val = self.params['log10_M0']
        if not hasattr(val, '__len__'):
            val = [val]
        return np.array(val)

    @property
    def log10_M1(self):
        """
        Log10 of the normalization mass for satellite galaxies.

        :type: array_like
        """
        val = self.params['log10_M1']
        if not hasattr(val, '__len__'):
            val = [val]
        return np.array(val)

    @property
    def alpha(self):
        """
        Slope parameter for satellite galaxies.

        :type: array_like
        """
        val = self.params['alpha']
        if not hasattr(val, '__len__'):
            val = [val]
        return np.array(val)[:, np.newaxis, np.newaxis]

    @property
    def sigma(self):
        """
        Scatter in the central galaxy occupation.

        :type: array_like
        """
        val = self.params['sigma']
        if not hasattr(val, '__len__'):
            val = [val]
        return np.array(val)[:, np.newaxis, np.newaxis]

    @property
    def Mmin(self):
        """
        Return the Mmin parameter in the correct shape and units.

        Returns:
        --------
        array_like
            Mmin
        """
        return 10.0 ** self.log10_Mmin[:, np.newaxis, np.newaxis]

    @property
    def M0(self):
        """
        Return the M0 parameter in the correct shape and units.

        Returns:
        --------
        array_like
            M0
        """
        return 10.0 ** self.log10_M0[:, np.newaxis, np.newaxis]

    @property
    def M1(self):
        """
        Return the M1 parameter in the correct shape and units.ž

        Returns:
        --------
        array_like
            M1
        """
        return 10.0 ** self.log10_M1[:, np.newaxis, np.newaxis]

    @property
    def _compute_hod_cen(self):
        """
        Compute the HOD for central galaxies.
        """
        N_cen = np.where(
            self.sigma == 0.0,
            np.heaviside(self.mass - self.Mmin, 1.0),
            0.5 * (1.0 + erf(np.log10(self.mass / self.Mmin) / self.sigma)),
        )
        if self.params['A_cen'] is not None:
            delta_pop_c = self.params['A_cen'] * np.fmin(N_cen, 1.0 - N_cen)
            N_cen = N_cen + delta_pop_c
        return np.tile(N_cen, (self.nz, 1))

    @property
    def _compute_hod_sat(self):
        """
        Compute the HOD for satellite galaxies.
        """
        N_sat = (
            np.heaviside(self.mass - self.M0, 1.0) * (self.mass - self.M0) / self.M1
        ) ** self.alpha
        if self.params['A_sat'] is not None:
            delta_pop_s = self.params['A_sat'] * N_sat
            N_sat = N_sat + delta_pop_s
        return np.tile(N_sat, (self.nz, 1))


class Zhai(HaloOccupationDistribution):
    """
    HOD model from Zhai et al. (2017) [1]_.

    Parameters:
    -----------
    log10_Mmin : array_like
        Log10 of the minimum mass for central galaxies.
    log10_Msat : array_like
        Log10 of the minimum mass for satellite galaxies.
    log10_Mcut : array_like
        Log10 of the cutoff mass for satellite galaxies.
    sigma : array_like
        Scatter in the central galaxy occupation.
    alpha : array_like
        Slope parameter for satellite galaxies.
    A_cen : float, optional
        Decorated HOD assembly bias parameter for central galaxies.
    A_sat : float, optional
        Decorated HOD assembly bias parameter for satellite galaxies.

    References
    ----------
    .. [1]  Zhai, Z. et al., "The clustering of luminous red galaxies at z ~ 0.7 from eBOSS and BOSS data",
            https://iopscience.iop.org/article/10.3847/1538-4357aa8eee.
    """

    _defaults = {
        'log10_Mmin': 13.68,
        'log10_Msat': 14.87,
        'log10_Mcut': 12.32,
        'sigma': 0.82,
        'alpha': 0.41,
        'A_cen': None,
        'A_sat': None,
    }

    @property
    def log10_Mmin(self):
        """
        Log10 of the minimum mass for central galaxies.

        :type: array_like
        """
        val = self.params['log10_Mmin']
        if not hasattr(val, '__len__'):
            val = [val]
        return np.array(val)

    @property
    def log10_Msat(self):
        """
        Log10 of the minimum mass for satellite galaxies.

        :type: array_like
        """
        val = self.params['log10_Msat']
        if not hasattr(val, '__len__'):
            val = [val]
        return np.array(val)

    @property
    def log10_Mcut(self):
        """
        Log10 of the cutoff mass for satellite galaxies.

        :type: array_like
        """
        val = self.params['log10_Mcut']
        if not hasattr(val, '__len__'):
            val = [val]
        return np.array(val)

    @property
    def alpha(self):
        """
        Slope parameter for satellite galaxies.

        :type: array_like
        """
        val = self.params['alpha']
        if not hasattr(val, '__len__'):
            val = [val]
        return np.array(val)[:, np.newaxis, np.newaxis]

    @property
    def sigma(self):
        """
        Scatter in the central galaxy occupation.

        :type: array_like
        """
        val = self.params['sigma']
        if not hasattr(val, '__len__'):
            val = [val]
        return np.array(val)[:, np.newaxis, np.newaxis]

    @property
    def Mmin(self):
        """
        Return the Mmin parameter in the correct shape and units.

        Returns:
        --------
        array_like
            Mmin
        """
        return 10.0 ** self.log10_Mmin[:, np.newaxis, np.newaxis]

    @property
    def Msat(self):
        """
        Return the Msat parameter in the correct shape and units.

        Returns:
        --------
        array_like
            Msat
        """
        return 10.0 ** self.log10_Msat[:, np.newaxis, np.newaxis]

    @property
    def Mcut(self):
        """
        Return the Mcut parameter in the correct shape and units.

        Returns:
        --------
        array_like
            Mcut
        """
        return 10.0 ** self.log10_Mcut[:, np.newaxis, np.newaxis]

    @property
    def _compute_hod_cen(self):
        """
        Compute the HOD for central galaxies.
        """
        N_cen = np.where(
            self.sigma == 0.0,
            np.heaviside(self.mass - self.Mmin, 1.0),
            0.5 * (1.0 + erf(np.log10(self.mass / self.Mmin) / self.sigma)),
        )
        if self.params['A_cen'] is not None:
            delta_pop_c = self.params['A_cen'] * np.fmin(N_cen, 1.0 - N_cen)
            N_cen = N_cen + delta_pop_c
        return np.tile(N_cen, (self.nz, 1))

    @property
    def _compute_hod_sat(self):
        """
        Compute the HOD for satellite galaxies.
        """
        # Paper has a Nc(M) multiplication, but I think the central condition covers this
        N_sat = ((self.mass / self.Msat) ** self.alpha) * np.exp(-self.Mcut / self.mass)
        if self.params['A_sat'] is not None:
            delta_pop_s = self.params['A_sat'] * N_sat
            N_sat = N_sat + delta_pop_s
        return np.tile(N_sat, (self.nz, 1))



import numpy as np

hod_params = {
    'log10_obs_norm_c': 9.95,
    'log10_m_ch': 11.24,
    'g1': 3.18,
    'g2': 0.245,
    'sigma_log10_O_c': 0.157,
    'norm_s': 0.562,
    'pivot': 12.0,
    'alpha_s': -1.18,
    'beta_s': 2,
    'b0': -1.17,
    'b1': 1.53,
    'b2': -0.217,
    'A_cen': None,
    'A_sat': None,
}

hod_settings = {
    'observables_file': None,
    'observable_z': False,
    'obs_min': np.array([8.0, 8.0]),
    'obs_max': np.array([12.0, 12.0]),
    'zmin': np.array([0.0, 0.2]),
    'zmax': np.array([0.2, 0.5]),
    'nz': 15,
    'nobs': 300,
    'observable_h_unit': '1/h^2',
}

hod_settings_mm = {
    'observables_file': '../input_files/red_cen_lum_pdf.txt',
    'observable_z': True,
}

obs_settings = {
    'observables_file': None,
    'observable_z': False,
    'obs_min': np.array([8.0]),
    'obs_max': np.array([12.0]),
    'zmin': np.array([0.1]),
    'zmax': np.array([0.1]),
    'nz': 1,
    'nobs': 300,
    'observable_h_unit': '1/h^2',
}

kwargs = {
    #        'matter_power_lin': None,
    #        'mead_correction': 'fit_feedback',
    #        'k_vec': np.logspace(-4, 4, 100),
    #        'z_vec': np.linspace(0.0, 3.0, 15),
    #        'lnk_min': np.log(10**(-4.0)),
    #        'lnk_max': np.log(10**(4.0)),
    #        'dlnk': (np.log(10**(-4.0)) - np.log(10**(4.0))) / 100,
    #        'dewiggle': True,
    'Mmin': 9.0,
    'Mmax': 18.0,
    'dlog10m': 0.005,
    #        'mdef_model': 'SOMean',
    #        'hmf_model': 'Tinker10',
    #        'bias_model': 'Tinker10',
    #        'transfer_model': 'CAMB',
    #        'transfer_params': {},
    #        'growth_model': 'CAMB',
    #        'growth_params': {},
    #        'norm_cen': 1.0,
    #        'norm_sat': 1.0,
    #        'eta_cen': 0.0,
    #        'eta_sat': 0.0,
    #        'overdensity': 200,
    #        'delta_c': 1.686,
    #        'one_halo_ktrunc': 0.1,
    #        'two_halo_ktrunc': 2.0,
    #        'matter_power_nl': None,
    #        'bnl': True,
    #        'beta_nl': None,
    #        'omega_c': 0.25,
    #        'omega_m': 0.3,
    #        'omega_b': 0.05,
    #        'h0': 0.7,
    #        'n_s': 0.9,
    #        'sigma_8': 0.8,
    #        'm_nu': 0.06,
    #        'w0': -1.0,
    #        'wa': 0.0,
    #        'tcmb': 2.7255,
    'log10T_AGN': 7.8,
    #        'mb': 10.0**13.0,
    'pointmass': True,
    'hod_model': 'Cacciato',
    'hod_params': hod_params,
    'hod_settings': hod_settings,
    'obs_settings': obs_settings,
    'compute_observable': True,
    'one_halo_ktrunc_ia': 4.0,
    'two_halo_ktrunc_ia': 6.0,
    'fortuna': False,
    't_eff': 0.0,
    'align_params': {
        'central_ia_depends_on': 'halo_mass',
        'satellite_ia_depends_on': 'halo_mass',
        'gamma_2h_amplitude': 5.33,
        'beta_cen': 0.44,
        'beta_two': None,
        'gamma_1h_amplitude': 0.0015,
        'gamma_1h_slope': -2.0,
        'beta_sat': 0.44,
        'mpivot_cen': 13.5,
        'mpivot_sat': 13.5,
        'nmass': 5,
        'n_hankel': 350,
        'nk': 10,
        'ell_max': 6,
    },
}



# A new power spectrum module

# NOTE: no truncation (halo exclusion problem) applied, as it is included in BNL!

import numpy as np
from collections import OrderedDict
from cosmosis.datablock import names, option_section
from dark_emulator import darkemu
from scipy.interpolate import RegularGridInterpolator, interp1d

#import sys
#sys.path.insert(0, "/net/home/fohlen13/dvornik/halo_model_mc/halomodel_for_cosmosis/package/pk")
#from bnl import NonLinearBias
from onepower.bnl import NonLinearBias

cosmo_params = names.cosmological_parameters

def setup(options):
    # This function is called once per processor per chain.
    # It is a chance to read any fixed options from the configuration file,
    # load any data, or do any calculations that are fixed once.

    log_mass_min = options[option_section, 'log_mass_min']
    log_mass_max = options[option_section, 'log_mass_max']
    nmass = options[option_section, 'nmass']
    dlog10m = (log_mass_max - log_mass_min) / nmass
    mass = 10 ** np.arange(log_mass_min, log_mass_max, dlog10m)

    # TODO: We might need to specify the mass binning of bnl, but for now it is not user accessible!
    # nmass_bnl = options[option_section, 'nmass_bnl']
    # mass_bnl = np.logspace(log_mass_min, log_mass_max, nmass_bnl)

    zmin = options[option_section, 'zmin']
    zmax = options[option_section, 'zmax']
    nz = options[option_section, 'nz']
    z_vec = np.linspace(zmin, zmax, nz)

    nk = options[option_section, 'nk']
    bnl = options.get_bool(option_section, 'bnl', default=False)

    if bnl:
        cached_bnl = {
            'num_calls': 0,
            'cached_bnl': None,
            'update_bnl': options[option_section, 'update_bnl']
        }
    else:
        cached_bnl = None

    return mass, nmass, z_vec, nz, nk, bnl, cached_bnl

def execute(block, config):
    # This function is called every time you have a new sample of cosmological and other parameters.
    # It is the main workhorse of the code. The block contains the parameters and results of any
    # earlier modules, and the config is what we loaded earlier.

    mass, nmass, z_vec, nz, nk, bnl, cached_bnl = config

    # load linear power spectrum
    k_vec_original = block['matter_power_lin', 'k_h']
    k_vec = np.logspace(np.log10(k_vec_original[0]), np.log10(k_vec_original[-1]), num=nk)

    if bnl:
        num_calls = cached_bnl['num_calls']
        update_bnl = cached_bnl['update_bnl']

        if num_calls % update_bnl == 0:
            bnl = NonLinearBias(
                mass = mass,
                z_vec = z_vec,
                k_vec = k_vec,
                h0 = block[cosmo_params, 'h0'],
                A_s = block[cosmo_params, 'A_s'],
                omega_b = block[cosmo_params, 'omega_b'],
                omega_c = block[cosmo_params, 'omega_c'],
                omega_lambda = 1.0 - block[cosmo_params, 'omega_m'],
                n_s = block[cosmo_params, 'n_s'],
                w0 = block[cosmo_params, 'w'],
                z_dep = False
            )

            beta_interp = bnl.bnl
            cached_bnl['cached_bnl'] = beta_interp
        else:
            beta_interp = cached_bnl['cached_bnl']

        cached_bnl['num_calls'] = num_calls + 1
        block.put_double_array_nd('bnl', 'beta_interp', beta_interp)
    else:
        block.put_double_array_nd('bnl', 'beta_interp', np.array([0.0]))

    return 0

def cleanup(config):
    # Usually python modules do not need to do anything here.
    # We just leave it in out of pedantic completeness.
    pass



import halo_model_utility as hmu
import halomod.concentration as concentration_classes
import halomod.profiles as profile_classes
import numpy as np
import time
import warnings
from astropy.cosmology import Flatw0waCDM
from cosmosis.datablock import names, option_section
from halomod.concentration import interp_concentration, make_colossus_cm
from halomod.halo_model import DMHaloModel

import hmf

# Silencing a warning from hmf for which the nonlinear mass is still correctly calculated
warnings.filterwarnings("ignore", message="Nonlinear mass outside mass range")

# Cosmological parameters section name in block
cosmo_params = names.cosmological_parameters

def setup(options):
    # Log10 Minimum, Maximum and number of log10 mass bins for halo masses: M_halo
    # Units are in log10(M_sun h^-1)
    log_mass_min = options[option_section, 'log_mass_min']
    log_mass_max = options[option_section, 'log_mass_max']
    nmass = options[option_section, 'nmass']
    dlog10m = (log_mass_max - log_mass_min) / nmass

    # Minimum and Maximum redshift and number of redshift bins for calculating the ingredients
    zmin = options[option_section, 'zmin']
    zmax = options[option_section, 'zmax']
    nz = options[option_section, 'nz']
    z_vec = np.linspace(zmin, zmax, nz)
    # If this is smaller than nz then downsample for concentration to speed it up.
    # The concentration function is slow!
    # nz_conc = options.get_int(option_section, 'nz_conc', default=5)

    # Model choices
    nk = options[option_section, 'nk']
    profile = options.get_string(option_section, 'profile', default='NFW')
    profile_value_name = options.get_string(option_section, 'profile_value_name', default='profile_parameters')
    hmf_model = options.get_string(option_section, 'hmf_model')
    mdef_model = options.get_string(option_section, 'mdef_model')
    overdensity = options[option_section, 'overdensity']
    cm_model = options.get_string(option_section, 'cm_model')
    delta_c = options[option_section, 'delta_c']
    bias_model = options.get_string(option_section, 'bias_model')
    mdef_params = {} if mdef_model == 'SOVirial' else {'overdensity': overdensity}

    # Option to set similar corrections to HMcode2020
    use_mead = options.get_string(option_section, 'use_mead2020_corrections', default='None')

    # Mapping of use_mead values to mead_correction values
    mead_correction_map = {
        'mead2020': 'nofeedback',
        'mead2020_feedback': 'feedback',
        # Add more mappings here if needed
    }

    # Determine the mead_correction based on the mapping
    mead_correction = mead_correction_map.get(use_mead, None)

    # If Mead correction is applied, set the ingredients to match Mead et al. (2021)
    if mead_correction is not None:
        hmf_model = 'ST'
        bias_model = 'ST99'
        mdef_model = 'SOVirial'
        mdef_params = {}
        cm_func = 'Duffy08'  # Dummy cm model, correct one calculated in execute
        disable_mass_conversion = True
    if mead_correction is None:
        try:
            cm_func = interp_concentration(getattr(concentration_classes, cm_model))
        except:
            cm_func = interp_concentration(make_colossus_cm(cm_model))
        disable_mass_conversion = False

    # Initialize cosmology model
    initialise_cosmo = Flatw0waCDM(H0=100.0, Ob0=0.044, Om0=0.3, Tcmb0=2.7255, w0=-1., wa=0.)

    # Growth Factor from hmf
    # gf._GrowthFactor.supported_cosmos = (FlatLambdaCDM, Flatw0waCDM, LambdaCDM)

    # Halo Mass function from hmf
    DM_hmf = DMHaloModel(
        z=0.0,
        cosmo_model=initialise_cosmo,
        sigma_8=0.8,
        n=0.96,
        transfer_model='CAMB',
        growth_model='CambGrowth',
        lnk_min=-18.0,
        lnk_max=18.0,
        dlnk=0.001,
        Mmin=log_mass_min,
        Mmax=log_mass_max,
        dlog10m=dlog10m,
        hmf_model=hmf_model,
        mdef_model=mdef_model,
        mdef_params=mdef_params,
        delta_c=delta_c,
        disable_mass_conversion=disable_mass_conversion,
        bias_model=bias_model,
        halo_profile_model=profile,
        halo_concentration_model=cm_func
    )

    # Array of halo masses
    mass = DM_hmf.m
    DM_hmf.ERROR_ON_BAD_MDEF = False

    # Configuration dictionary
    config = {
        'z_vec': z_vec,
        'nz': nz,
        'mass': mass,
        'DM_hmf': DM_hmf,
        'mdef_model': mdef_model,
        'overdensity': overdensity,
        'delta_c': delta_c,
        'mead_correction': mead_correction,
        'nk': nk,
        'profile_value_name': profile_value_name
    }

    return config

def execute(block, config):
    # Read in the config as returned by setup
    z_vec = config['z_vec']
    nz = config['nz']
    mass = config['mass']
    DM_hmf = config['DM_hmf']
    mdef_model = config['mdef_model']
    overdensity = config['overdensity']
    delta_c = config['delta_c']
    mead_correction = config['mead_correction']
    nk = config['nk']
    profile_value_name = config['profile_value_name']

    # Astropy cosmology requires the CMB temperature as an input.
    # If it exists in the values file read it from there otherwise set to its default value
    tcmb = block.get_double(cosmo_params, 'TCMB', default=2.7255)

    # Update the cosmological parameters
    this_cosmo_run = Flatw0waCDM(
        H0=block[cosmo_params, 'hubble'],
        Ob0=block[cosmo_params, 'omega_b'],
        Om0=block[cosmo_params, 'omega_m'],
        m_nu=[0, 0, block[cosmo_params, 'mnu']],
        Tcmb0=tcmb,
        w0=block[cosmo_params, 'w'],
        wa=block[cosmo_params, 'wa']
    )
    ns = block[cosmo_params, 'n_s']
    sigma_8 = block[cosmo_params, 'sigma_8']

    # TODO: will the inputs depend on the profile model?
    norm_cen = block[profile_value_name, 'norm_cen']
    norm_sat = block[profile_value_name, 'norm_sat']
    eta_cen = block[profile_value_name, 'eta_cen']
    eta_sat = block[profile_value_name, 'eta_sat']

    # Initialize arrays
    nmass_hmf = len(mass)
    dndlnmh = np.empty([nz, nmass_hmf])
    nu = np.empty([nz, nmass_hmf])
    b_nu = np.empty([nz, nmass_hmf])
    rho_halo = np.empty([nz])
    neff = np.empty([nz])
    sigma8_z = np.empty([nz])
    f_nu = np.empty([nz])
    mean_density_z = np.empty([nz])
    halo_overdensity_mean = np.empty([nz])
    u_dm_cen = np.empty([nz, nk, nmass_hmf])
    u_dm_sat = np.empty([nz, nk, nmass_hmf])
    conc_cen = np.empty([nz, nmass_hmf])
    conc_sat = np.empty([nz, nmass_hmf])
    r_s_cen = np.empty([nz, nmass_hmf])
    r_s_sat = np.empty([nz, nmass_hmf])
    rvir_cen = np.empty([nz, nmass_hmf])
    rvir_sat = np.empty([nz, nmass_hmf])

    if mead_correction:
        growth = hmu.get_growth_interpolator(this_cosmo_run)
        # growth_LCDM = hmu.get_growth_interpolator(LCDMcosmo)

    # Get the k range from the linear matter power spectrum section.
    # But use the nk that was given as input
    k_vec_original = block['matter_power_lin', 'k_h']
    k = np.logspace(np.log10(k_vec_original[0]), np.log10(k_vec_original[-1]), num=nk)

    # Power spectrum transfer function used to update the transfer function in hmf
    transfer_k = block['matter_power_transfer_func', 'k_h']
    transfer_func = block['matter_power_transfer_func', 't_k']
    growth_z = block['growth_parameters', 'z']
    growth_func = block['growth_parameters', 'd_z']

    DM_hmf.update(
        cosmo_model=this_cosmo_run,
        sigma_8=sigma_8,
        n=ns,
        transfer_model='FromArray',
        transfer_params={'k': transfer_k, 'T': transfer_func},
        halo_profile_params={'cosmo': this_cosmo_run},
        growth_model='FromArray',
        growth_params={'z': growth_z, 'd': growth_func}
    )

    # Loop over a series of redshift values defined by z_vec = np.linspace(zmin, zmax, nz)
    for jz, z_iter in enumerate(z_vec):
        if mead_correction is not None:
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=UserWarning)
                # This disables the warning from hmf. hmf is just telling us what we know
                # hmf's internal way of calculating the overdensity and the collapse threshold are fixed.
                # When we use the mead correction we want to define the haloes using the virial definition.
                # To avoid conflicts we manually pass the overdensity and the collapse threshold,
                # but for that we need to set the mass definition to be "mean",
                # so that it is compared to the mean density of the Universe rather than critical density.
                # hmf warns us that the value is not a native definition for the given halo mass function,
                # but will interpolate between the known ones (this is happening when one uses Tinker hmf for instance).
                a = this_cosmo_run.scale_factor(z_iter)
                g = growth(a)
                G = hmu.get_accumulated_growth(a, growth)
                delta_c_z = hmu.dc_Mead(a, this_cosmo_run.Om(z_iter) + this_cosmo_run.Onu(z_iter),
                                        this_cosmo_run.Onu0 / (this_cosmo_run.Om0 + this_cosmo_run.Onu0), g, G)
                halo_overdensity_mead = hmu.Dv_Mead(a, this_cosmo_run.Om(z_iter) + this_cosmo_run.Onu(z_iter),
                                                    this_cosmo_run.Onu0 / (this_cosmo_run.Om0 + this_cosmo_run.Onu0), g, G)
                DM_hmf.update(
                    z=z_iter,
                    delta_c=delta_c_z,
                    mdef_model=hmu.SOVirial_Mead,
                    mdef_params={'overdensity': halo_overdensity_mead}
                )

                eta_cen = 0.1281 * DM_hmf.sigma8_z**(-0.3644)
                if mead_correction == 'nofeedback':
                    norm_cen = 5.196  # /3.85#1.0#(5.196/3.85) #0.85*1.299
                elif mead_correction == 'feedback':
                    theta_agn = block['halo_model_parameters', 'logT_AGN'] - 7.8
                    norm_cen = (5.196 / 4.0) * ((3.44 - 0.496 * theta_agn) * np.power(10.0, z_iter * (-0.0671 - 0.0371 * theta_agn)))

                zf = hmu.get_halo_collapse_redshifts(mass, z_iter, delta_c_z, growth, this_cosmo_run, DM_hmf)
                conc_cen[jz, :] = norm_cen * (1.0 + zf) / (1.0 + z_iter)
                conc_sat[jz, :] = norm_sat * (1.0 + zf) / (1.0 + z_iter)

                DM_hmf.update(halo_profile_params={'eta_bloat': eta_cen})
                nfw_cen = DM_hmf.halo_profile.u(k, DM_hmf.m, c=conc_cen[jz, :])
                u_dm_cen[jz, :, :] = nfw_cen / np.expand_dims(nfw_cen[0, :], 0)
                r_s_cen[jz, :] = DM_hmf.halo_profile._rs_from_m(DM_hmf.m)
                rvir_cen[jz, :] = DM_hmf.halo_profile.halo_mass_to_radius(DM_hmf.m)

                DM_hmf.update(halo_profile_params={'eta_bloat': eta_sat})
                nfw_sat = DM_hmf.halo_profile.u(k, DM_hmf.m, c=conc_sat[jz, :])
                u_dm_sat[jz, :, :] = nfw_sat / np.expand_dims(nfw_sat[0, :], 0)
                r_s_sat[jz, :] = DM_hmf.halo_profile._rs_from_m(DM_hmf.m)
                rvir_sat[jz, :] = DM_hmf.halo_profile.halo_mass_to_radius(DM_hmf.m)

        if mead_correction is None:
            delta_c_z = (
                (3.0 / 20.0) * (12.0 * np.pi) ** (2.0 / 3.0) * (1.0 + 0.0123 * np.log10(this_cosmo_run.Om(z_iter)))
                if mdef_model == 'SOVirial' else
                delta_c
            )

            DM_hmf.update(
                z=z_iter,
                delta_c=delta_c_z,
                halo_profile_params={'eta_bloat': eta_cen},
                halo_concentration_params={'norm': norm_cen}
            )

            conc_cen[jz, :] = DM_hmf.cmz_relation
            nfw_cen = DM_hmf.halo_profile.u(k, DM_hmf.m)
            u_dm_cen[jz, :, :] = nfw_cen / np.expand_dims(nfw_cen[0, :], 0)
            r_s_cen[jz, :] = DM_hmf.halo_profile._rs_from_m(DM_hmf.m)
            rvir_cen[jz, :] = DM_hmf.halo_profile.halo_mass_to_radius(DM_hmf.m)

            DM_hmf.update(halo_profile_params={'eta_bloat': eta_sat},
                          halo_concentration_params={'norm': norm_sat})
            conc_sat[jz, :] = DM_hmf.cmz_relation
            nfw_sat = DM_hmf.halo_profile.u(k, DM_hmf.m)
            u_dm_sat[jz, :, :] = nfw_sat / np.expand_dims(nfw_sat[0, :], 0)
            r_s_sat[jz, :] = DM_hmf.halo_profile._rs_from_m(DM_hmf.m)
            rvir_sat[jz, :] = DM_hmf.halo_profile.halo_mass_to_radius(DM_hmf.m)

        halo_overdensity_mean[jz] = DM_hmf.halo_overdensity_mean
        nu[jz, :] = DM_hmf.nu**0.5
        dndlnmh[jz, :] = DM_hmf.dndlnm
        mean_density0 = DM_hmf.mean_density0
        mean_density_z[jz] = DM_hmf.mean_density
        rho_halo[jz] = halo_overdensity_mean[jz] * DM_hmf.mean_density0
        b_nu[jz, :] = DM_hmf.halo_bias

        neff[jz] = DM_hmf.n_eff_at_collapse
        # Rnl = DM_hmf.filter.mass_to_radius(DM_hmf.mass_nonlinear, DM_hmf.mean_density0)
        # neff[jz] = -3.0 - 2.0*DM_hmf.normalised_filter.dlnss_dlnm(Rnl)

        # Only used for mead_corrections
        sigma8_z[jz] = DM_hmf.sigma8_z
        # pk_cold = DM_hmf.power * hmu.Tk_cold_ratio(DM_hmf.k, g, block[cosmo_params, 'ommh2'], block[cosmo_params, 'h0'], this_cosmo_run.Onu0/this_cosmo_run.Om0, this_cosmo_run.Neff, T_CMB=tcmb)**2.0
        # sigma8_z[jz] = hmu.sigmaR_cc(pk_cold, DM_hmf.k, 8.0)

    # TODO: Clean these up. Put more of them into the same folder
    block.put_grid('concentration_m', 'z', z_vec, 'm_h', mass, 'c', conc_cen)
    block.put_grid('concentration_sat', 'z', z_vec, 'm_h', mass, 'c', conc_sat)
    block.put_grid('nfw_scale_radius_m', 'z', z_vec, 'm_h', mass, 'rs', r_s_cen)
    block.put_grid('nfw_scale_radius_sat', 'z', z_vec, 'm_h', mass, 'rs', r_s_sat)

    block.put_double_array_1d('virial_radius', 'm_h', mass)
    # rvir doesn't change with z, hence no z-dimension
    block.put_double_array_1d('virial_radius', 'rvir_m', rvir_cen[0])
    block.put_double_array_1d('virial_radius', 'rvir_sat', rvir_sat[0])

    block.put_double_array_1d('fourier_nfw_profile', 'z', z_vec)
    block.put_double_array_1d('fourier_nfw_profile', 'm_h', mass)
    block.put_double_array_1d('fourier_nfw_profile', 'k_h', k)
    block.put_double_array_nd('fourier_nfw_profile', 'ukm', u_dm_cen)
    block.put_double_array_nd('fourier_nfw_profile', 'uksat', u_dm_sat)

    ###################################################################################################################

    # Density
    block['density', 'mean_density0'] = mean_density0
    block['density', 'rho_crit'] = mean_density0 / this_cosmo_run.Om0
    block.put_double_array_1d('density', 'mean_density_z', mean_density_z)
    block.put_double_array_1d('density', 'rho_halo', rho_halo)
    block.put_double_array_1d('density', 'z', z_vec)

    # Halo mass function
    block.put_grid('hmf', 'z', z_vec, 'm_h', mass, 'dndlnmh', dndlnmh)
    block.put_grid('hmf', 'z', z_vec, 'm_h', mass, 'nu', nu)
    block.put_double_array_1d('hmf', 'neff', neff)
    block.put_double_array_1d('hmf', 'sigma8_z', sigma8_z)

    # Linear halo bias
    block.put_grid('halobias', 'z', z_vec, 'm_h', mass, 'b_hb', b_nu)

    # Fraction of neutrinos to total matter, f_nu = Ω_nu /Ω_m
    f_nu = this_cosmo_run.Onu0 / this_cosmo_run.Om0
    block[cosmo_params, 'fnu'] = f_nu

    config['DM_hmf'] = DM_hmf

    return 0

def cleanup(config):
    # Usually python modules do not need to do anything here.
    # We just leave it in out of pedantic completeness.
    pass



##############################
    ##  wrapper_twopoint.py    ##
    ##  Chieh-An Lin        ##
    ##  Version 2020.02.21    ##
    ##############################

import collections as clt
import numpy as np
import twopoint

###############################################################################
## Functions related to label convention

class LabelConvention:
    """
    There can be 3 different names for 1 statistics.
    Say for example, xi_+:
    In the theory pipeline, it is something like "shear_xi_plus"
    In the scale cuts module, it's "xiP" (cannot have underscore for the string-splitting reason).
    In the measurement pipeline, it depends on user's choice.
    This class deals with the conversion.
    """

    def __init__(self, w='wtheta', gamma_t='gammat', gamma_x='gammax', xi_p='xip', xi_m='xim',
                  P_nn='Pnn', P_ne_E='PneE', P_ne_B='PneB', P_ee_E='PeeE', P_ee_B='PeeB',
                  E_n='En', B_n='Bn', Psi_gm='Psi_gm', Psi_gg='Psi_gg', onept='1pt'):

        self.prefix  = 'twoPoint'
        self.lens    = 'NZ_LENS'
        self.source  = 'NZ_SOURCE'
        self.obs     = 'NZ_OBS'

        self.onept   = '1PT'.lower()

        self.w       = 'wtheta'.lower()
        self.gamma_t = 'gammat'.lower()
        self.gamma_x = 'gammax'.lower()
        self.xi_p    = 'xip'.lower()
        self.xi_m    = 'xim'.lower()
        self.P_nn    = 'Pnn'.lower()
        self.P_ne_E  = 'PneE'.lower()
        self.P_ne_B  = 'PneB'.lower()
        self.P_ee_E  = 'PeeE'.lower()
        self.P_ee_B  = 'PeeB'.lower()
        self.E_n     = 'En'.lower()
        self.B_n     = 'Bn'.lower()
        self.Psi_gm  = 'Psi_gm'.lower()
        self.Psi_gg  = 'Psi_gg'.lower()

        self.labDict = {
            self.w:        w,
            self.gamma_t:  gamma_t,
            self.gamma_x:  gamma_x,
            self.xi_p:     xi_p,
            self.xi_m:     xi_m,
            self.P_nn:     P_nn,
            self.P_ne_E:   P_ne_E,
            self.P_ne_B:   P_ne_B,
            self.P_ee_E:   P_ee_E,
            self.P_ee_B:   P_ee_B,
            self.E_n:      E_n,
            self.B_n:      B_n,
            self.Psi_gm:   Psi_gm,
            self.Psi_gg:   Psi_gg,
            self.onept:    onept,
            'all':         'all'
        }

        self.revLabDict = {}
        for key, value in self.labDict.items():
            self.revLabDict[value] = key

        tpType1 = twopoint.Types.galaxy_position_real       ## GPR
        tpType2 = twopoint.Types.galaxy_shear_plus_real     ## G+R
        tpType3 = twopoint.Types.galaxy_shear_minus_real    ## G-R
        tpType4 = twopoint.Types.galaxy_position_fourier    ## GPF
        tpType5 = twopoint.Types.galaxy_shear_emode_fourier ## GEF
        tpType6 = twopoint.Types.galaxy_shear_bmode_fourier ## GBF

        self.kernelTypeDict = clt.OrderedDict()
        ## Don't touch the order of the assertion
        self.kernelTypeDict[w]       = [self.lens,   self.lens,   tpType1, tpType1, 'arcmin']
        self.kernelTypeDict[gamma_t] = [self.lens,   self.source, tpType1, tpType2, 'arcmin']
        self.kernelTypeDict[gamma_x] = [self.lens,   self.source, tpType1, tpType3, 'arcmin']
        self.kernelTypeDict[xi_p]    = [self.source, self.source, tpType2, tpType2, 'arcmin']
        self.kernelTypeDict[xi_m]    = [self.source, self.source, tpType3, tpType3, 'arcmin']
        self.kernelTypeDict[P_nn]    = [self.lens,   self.lens,   tpType4, tpType4, None]
        self.kernelTypeDict[P_ne_E]  = [self.lens,   self.source, tpType4, tpType5, None]
        self.kernelTypeDict[P_ne_B]  = [self.lens,   self.source, tpType4, tpType6, None]
        self.kernelTypeDict[P_ee_E]  = [self.source, self.source, tpType5, tpType5, None]
        self.kernelTypeDict[P_ee_B]  = [self.source, self.source, tpType6, tpType6, None]
        self.kernelTypeDict[E_n]     = [self.source, self.source, tpType5, tpType5, None]
        self.kernelTypeDict[B_n]     = [self.source, self.source, tpType6, tpType6, None]
        self.kernelTypeDict[Psi_gm]  = [self.lens,   self.source, tpType4, tpType5, None]
        self.kernelTypeDict[Psi_gg]  = [self.lens,   self.lens,   tpType4, tpType4, None]
        return

    def defaultToCustomStatsTag(self, statsTag):
        statsList   = statsTag.split('+')
        statsList_c = [self.labDict[stats.lower()] for stats in statsList]
        statsTag_c  = '+'.join(statsList_c)
        return statsTag_c

    def customToDefaultStatsTag(self, statsTag_c):
        statsList_c = statsTag_c.split('+')
        statsList   = [self.revLabDict[stats_c] for stats_c in statsList_c]
        statsTag    = '+'.join(statsList)
        return statsTag

    def makeScaleCutsArgs(self, scDict):
        """
        Convert some lines of ASCII characters into a tuple of 5 elements defining various scale cuts
        scDict is a dictionary, constructed by:
        scDict[scale_cut_option] = list of strings
        """
        statsList = None
        cutCross = False
        statsTag_tomoInd_tomoInd_list = []
        statsTag_binIndList_dict = {}
        statsTag_tomoInd1_tomoInd2__angMin_angMax_dict = {}
        statsTag__angMin_angMax_dict = {}

        for key, value in scDict.items():
            keySplit = key.split('_')

            if key == 'use_stats':
                statsList = value.split()

            elif key == 'cut_cross':
                cutCross = bool(value)

            elif 'cut_pair' in key:
                statsTag = self.labDict[keySplit[2]]
                for pair in value.split():
                    pair = pair.split('+')
                    tomoInd1 = int(pair[0])
                    tomoInd2 = int(pair[1])
                    statsTag_tomoInd_tomoInd_list.append((statsTag, tomoInd1, tomoInd2))

            elif 'cut_bin' in key:
                statsTag   = self.labDict[keySplit[2]]
                binIndList = [int(ind) for ind in value.split()]
                statsTag_binIndList_dict[statsTag] = binIndList

            elif 'keep_ang' in key and len(keySplit) > 4:
                statsTag = self.labDict[keySplit[2]]
                tomoInd1 = int(keySplit[3])
                tomoInd2 = int(keySplit[4])
                value    = value.split()
                angMin   = float(value[0])
                angMax   = float(value[1])
                statsTag_tomoInd1_tomoInd2__angMin_angMax_dict[(statsTag, tomoInd1, tomoInd2)] = (angMin, angMax)

            elif 'keep_ang' in key and len(keySplit) < 4:
                statsTag = self.labDict[keySplit[2]]
                value    = value.split()
                angMin   = float(value[0])
                angMax   = float(value[1])
                statsTag__angMin_angMax_dict[statsTag] = (angMin, angMax)

        scArgs = cutCross, statsTag_tomoInd_tomoInd_list, statsTag_binIndList_dict, \
            statsTag_tomoInd1_tomoInd2__angMin_angMax_dict, statsTag__angMin_angMax_dict
        return statsList, scArgs

###############################################################################
## Functions related to SpectrumMeasurement builder

class SpectrumBuilder():

    def __init__(self):
        self.tIL1    = []
        self.tIL2    = []
        self.aIL     = []
        self.angList = []
        self.valList = []
        self.nobs = []
        self.obs = []
        self.obs_name = []
        self.nbin = []
        return

    def addTomo(self, tomoInd1, tomoInd2, angle, value):
        N_ang  = len(angle)
        angInd = np.arange(N_ang, dtype=int)
        self.tIL1.append([tomoInd1+1] * N_ang)
        self.tIL2.append([tomoInd2+1] * N_ang)
        self.aIL.append(angInd + 1)
        self.angList.append(angle)
        self.valList.append(value)
        return

    def add_one_point(self, obs_name, nbin, x, angbin, values):
        self.obs_name.append(obs_name)
        self.nbin.append(nbin)
        self.obs.append(x)
        self.nobs.append(values)
        #self.angbin.append(angbin)

    def makeSpectrum(self, name, types, angle_unit, kernels=(None, None)):
        self.tIL1    = np.concatenate(self.tIL1)
        self.tIL2    = np.concatenate(self.tIL2)
        self.aIL     = np.concatenate(self.aIL)
        self.angList = np.concatenate(self.angList)
        self.valList = np.concatenate(self.valList)
        spec = twopoint.SpectrumMeasurement(name, (self.tIL1, self.tIL2), types, kernels, 'SAMPLE', self.aIL, self.valList, angle=self.angList, angle_unit=angle_unit)
        return spec

###############################################################################
## Functions related to TwoPointFile wrapper

class TwoPointWrapper(twopoint.TwoPointFile):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        ## Default arguments are: spectra, kernels, nobs, windows, covmat_info
        return

    @classmethod
    def from_spectra(cls, spectra, kernels=None, nobs=None, covmat_info=None):
        TP = cls(spectra, kernels, nobs, 'SAMPLE', covmat_info) ## windows = 'SAMPLE'
        return TP

    @classmethod
    def from_fits(cls, filename, covmat_name=None):
        TP = super().from_fits(filename, covmat_name=covmat_name)
        return TP

    ## to_fits(self, filename, overwrite=False, clobber=False) can be used directly as
    ## TP.to_fits(filename, overwrite=True)

    def makeMeanVector(self):
        theory = []
        for spectrum in self.spectra:
            theory.append(spectrum.value)
        if self.nobs is not None:
            for obs in self.nobs:
                theory.append(np.array(obs.nobs).flatten())
        theory = np.concatenate(theory)
        return theory

    def replaceMeanVector(self, data):
        idx = []
        for spectrum in self.spectra:
            idx.append(len(spectrum.value))
        if self.nobs is not None:
            for obs in self.nobs:
                idx.append(len(np.array(obs.nobs).flatten()))

        if sum(idx) != len(data):
            raise ValueError(f"Size of provided data vector incompatible with spectra: {len(data)} vs {sum(idx)}.")

        idx.insert(0, 0)
        idx = np.cumsum(idx)

        for i, spectrum in enumerate(self.spectra):
            spectrum.value = data[idx[i]:idx[i+1]]
        if self.nobs is not None:
            for i, obs in enumerate(self.nobs):
                obs.nobs[i] = np.array(data[idx[len(self.spectra)+i]:idx[len(self.spectra)+i+1]]).flatten()

    def cutScales(self, cutCross=False, statsTag_tomoInd_tomoInd_list=[], statsTag_binIndList_dict={}):
        if cutCross:
            self.mask_cross() ## Cut cross pair bins, but not cross-corr between auto pair bins; e.g. C_12_12 masked, C_11_22 preserved

        if len(statsTag_tomoInd_tomoInd_list) > 0:
            self.mask_scales(bin_cuts=statsTag_tomoInd_tomoInd_list) ## Cut the whole tomo pair bin

        for statsTag, binIndList in statsTag_binIndList_dict.items():
            self.mask_indices(statsTag, binIndList) ## Cut by bin indices, 0-indexing
        return

    def keepScales(self, statsTag_tomoInd1_tomoInd2__angMin_angMax_dict={}, statsTag__angMin_angMax_dict={}):
        if len(statsTag_tomoInd1_tomoInd2__angMin_angMax_dict) > 0:
            self.mask_scales(cuts=statsTag_tomoInd1_tomoInd2__angMin_angMax_dict) ## Keep an angular range for specific tomo bins

        for statsTag, angMin_angMax in statsTag__angMin_angMax_dict.items():
            self.mask_scale(statsTag, min_scale=angMin_angMax[0], max_scale=angMin_angMax[1]) ## Keep an angular range for a specific tomo bin
        return

###############################################################################



import halo_model_utility as hmu
import halomod.concentration as concentration_classes
import halomod.profiles as profile_classes
import numpy as np
import time
import warnings
from astropy.cosmology import Flatw0waCDM
from cosmosis.datablock import names, option_section
from halomod.concentration import interp_concentration, make_colossus_cm
from halomod.halo_model import DMHaloModel

import hmf
from onepower.hmi import HaloModelIngredients

# Silencing a warning from hmf for which the nonlinear mass is still correctly calculated
warnings.filterwarnings("ignore", message="Nonlinear mass outside mass range")

# Cosmological parameters section name in block
cosmo_params = names.cosmological_parameters

def setup(options):

    config = {}
    # Log10 Minimum, Maximum and number of log10 mass bins for halo masses: M_halo
    # Units are in log10(M_sun h^-1)
    config['log_mass_min'] = options[option_section, 'log_mass_min']
    config['log_mass_max'] = options[option_section, 'log_mass_max']
    nmass = options[option_section, 'nmass']
    config['dlog10m'] = ( config['log_mass_max'] - config['log_mass_min']) / nmass

    # Minimum and Maximum redshift and number of redshift bins for calculating the ingredients
    zmin = options[option_section, 'zmin']
    zmax = options[option_section, 'zmax']
    nz = options[option_section, 'nz']
    config['z_vec'] = np.linspace(zmin, zmax, nz)
    # If this is smaller than nz then downsample for concentration to speed it up.
    # The concentration function is slow!
    # nz_conc = options.get_int(option_section, 'nz_conc', default=5)

    # Model choices
    config['nk'] = options[option_section, 'nk']
    config['profile'] = options.get_string(option_section, 'profile', default='NFW')
    config['profile_value_name'] = options.get_string(option_section, 'profile_value_name', default='profile_parameters')
    config['hmf_model'] = options.get_string(option_section, 'hmf_model')
    config['mdef_model'] = options.get_string(option_section, 'mdef_model')
    config['overdensity'] = options[option_section, 'overdensity']
    config['cm_model'] = options.get_string(option_section, 'cm_model')
    config['delta_c'] = options[option_section, 'delta_c']
    config['bias_model'] = options.get_string(option_section, 'bias_model')

    # Option to set similar corrections to HMcode2020
    use_mead = options.get_string(option_section, 'use_mead2020_corrections', default='None')

    # Mapping of use_mead values to mead_correction values
    mead_correction_map = {
        'mead2020': 'nofeedback',
        'mead2020_feedback': 'feedback',
        # Add more mappings here if needed
    }

    # Determine the mead_correction based on the mapping
    config['mead_correction'] = mead_correction_map.get(use_mead, None)

    config['lnk_min'] = -18.0
    config['lnk_max'] = 18.0
    config['dlnk'] = 0.001

    return config

def execute(block, config):


    # TODO: will the inputs depend on the profile model?
    norm_cen = block[config['profile_value_name'], 'norm_cen']
    norm_sat = block[config['profile_value_name'], 'norm_sat']
    eta_cen = block[config['profile_value_name'], 'eta_cen']
    eta_sat = block[config['profile_value_name'], 'eta_sat']

    # Get the k range from the linear matter power spectrum section.
    # But use the nk that was given as input
    k_vec_original = block['matter_power_lin', 'k_h']
    k_vec = np.logspace(np.log10(k_vec_original[0]), np.log10(k_vec_original[-1]), num=config['nk'])

    # Power spectrum transfer function used to update the transfer function in hmf
    transfer_k = block['matter_power_transfer_func', 'k_h']
    transfer_func = block['matter_power_transfer_func', 't_k']
    growth_z = block['growth_parameters', 'z']
    growth_func = block['growth_parameters', 'd_z']

    z_vec = config['z_vec']

    hmf = HaloModelIngredients(
        k_vec = k_vec,
        z_vec = z_vec,
        lnk_min = config['lnk_min'],
        lnk_max = config['lnk_max'],
        dlnk = config['dlnk'],
        Mmin = config['log_mass_min'],
        Mmax = config['log_mass_max'],
        dlog10m = config['dlog10m'],
        mdef_model = config['mdef_model'],
        hmf_model = config['hmf_model'],
        bias_model = config['bias_model'],
        halo_profile_model_dm = config['profile'],
        halo_concentration_model_dm = config['cm_model'],
        halo_profile_model_sat = config['profile'],
        halo_concentration_model_sat = config['cm_model'],
        transfer_model = 'FromArray',
        transfer_params = {'k': transfer_k, 'T': transfer_func},
        growth_model = 'FromArray',
        growth_params = {'z': growth_z, 'd': growth_func},
        h0 = block[cosmo_params, 'h0'],
        omega_c = block[cosmo_params, 'omega_c'],
        omega_b = block[cosmo_params, 'omega_b'],
        omega_m = block[cosmo_params, 'omega_m'],
        w0 = block[cosmo_params, 'w'],
        wa = block[cosmo_params, 'wa'],
        n_s = block[cosmo_params, 'n_s'],
        tcmb = block.get_double(cosmo_params, 'TCMB', default=2.7255),
        m_nu = block[cosmo_params, 'mnu'],
        sigma_8 = block[cosmo_params, 'sigma_8'],
        log10T_AGN = block['halo_model_parameters', 'logT_AGN'],
        norm_cen = norm_cen,
        norm_sat = norm_sat,
        eta_cen = eta_cen,
        eta_sat = eta_sat,
        overdensity = config['overdensity'],
        delta_c = config['delta_c'],
        mead_correction = config['mead_correction']
    )

    mass = hmf.mass

    u_dm_cen = hmf.u_dm
    u_dm_sat = hmf.u_sat
    mean_density0 = hmf.mean_density0
    mean_density_z = hmf.mean_density_z
    rho_crit = hmf.mean_density0 / block[cosmo_params, 'omega_m']
    rho_halo = hmf.rho_halo

    dndlnm = hmf.dndlnm
    halo_bias = hmf.halo_bias
    nu = hmf.nu
    neff = hmf.neff
    sigma8_z = hmf.sigma8_z
    fnu = hmf.fnu

    conc_cen = hmf.conc_cen
    conc_sat = hmf.conc_sat
    r_s_cen = hmf.r_s_cen
    r_s_sat = hmf.r_s_sat

    rvir_cen = hmf.rvir_cen
    rvir_sat = hmf.rvir_sat

    # TODO: Clean these up. Put more of them into the same folder
    block.put_grid('concentration_m', 'z', z_vec, 'm_h', mass, 'c', conc_cen)
    block.put_grid('concentration_sat', 'z', z_vec, 'm_h', mass, 'c', conc_sat)
    block.put_grid('nfw_scale_radius_m', 'z', z_vec, 'm_h', mass, 'rs', r_s_cen)
    block.put_grid('nfw_scale_radius_sat', 'z', z_vec, 'm_h', mass, 'rs', r_s_sat)

    block.put_double_array_1d('virial_radius', 'm_h', mass)
    # rvir doesn't change with z, hence no z-dimension
    block.put_double_array_1d('virial_radius', 'rvir_m', rvir_cen[0])
    block.put_double_array_1d('virial_radius', 'rvir_sat', rvir_sat[0])

    block.put_double_array_1d('fourier_nfw_profile', 'z', z_vec)
    block.put_double_array_1d('fourier_nfw_profile', 'm_h', mass)
    block.put_double_array_1d('fourier_nfw_profile', 'k_h', k_vec)
    block.put_double_array_nd('fourier_nfw_profile', 'ukm', u_dm_cen)
    block.put_double_array_nd('fourier_nfw_profile', 'uksat', u_dm_sat)


    # Density
    block['density', 'mean_density0'] = mean_density0
    block['density', 'rho_crit'] = rho_crit
    block.put_double_array_1d('density', 'mean_density_z', mean_density_z)
    block.put_double_array_1d('density', 'rho_halo', rho_halo)
    block.put_double_array_1d('density', 'z', z_vec)

    # Halo mass function
    block.put_grid('hmf', 'z', z_vec, 'm_h', mass, 'dndlnmh', dndlnm)
    block.put_grid('hmf', 'z', z_vec, 'm_h', mass, 'nu', nu)
    block.put_double_array_1d('hmf', 'neff', neff)
    block.put_double_array_1d('hmf', 'sigma8_z', sigma8_z)

    # Linear halo bias
    block.put_grid('halobias', 'z', z_vec, 'm_h', mass, 'b_hb', halo_bias)

    # Fraction of neutrinos to total matter, f_nu = Ω_nu /Ω_m
    block[cosmo_params, 'fnu'] = fnu

    return 0

def cleanup(config):
    # Usually python modules do not need to do anything here.
    # We just leave it in out of pedantic completeness.
    pass



import numpy as np
from functools import cached_property
from scipy.interpolate import interp1d

from hmf._internals._cache import parameter
from hmf._internals._framework import Framework

from .pk import PowerSpectrumResult, Spectra


class UpsampledSpectra(Framework):
    """
    This class generates one or two :class:`~halomodel.pk.Spectra`,
    extrapolates them to the desired output grid and adds them if requested.

    Parameters
    ----------
    z : array_like
        Output redshifts to which to interpolate and optionally extrapolate power spectra.
    k : array_like
        Output k-vector to which to interpolate and optionally extrapolate power spectra.
    fraction_z : array_like
        Redshifts of the red/blue fraction of galaxies in sample.
    fraciton : array_like
        Red/blue fraction of galaxies in sample as a function of redshift.
    model_1_params : dict
        Parameters for the first halo model.
    model_2_params : dict
        Parameters for the second halo model.
    extrapolate_option : str or float
        Extrapolation option to pass to interp1d.
    model : object
        Instance of Spectra(), pre-initialised for saving computing resources.
    """

    def __init__(
        self,
        z=0.0,
        k=0.0,
        fraction_z=None,
        fraction=None,
        model=None,
        model_1_params: dict | None = None,
        model_2_params: dict | None = None,
        extrapolate_option='extrapolate',
    ):
        super().__init__()
        self.z = z
        self.k = k
        self.fraction_z = fraction_z
        self.fraction = fraction
        self.model = model
        self._model_1_params = model_1_params or {}
        self._model_2_params = model_2_params
        self.extrapolate_option = extrapolate_option

    @parameter('param')
    def model(self, val):
        """
        Instance of Spectra(), pre-initialised for saving computing resources.

        :type: object
        """
        return val

    @parameter('param')
    def _model_1_params(self, val):
        """
        Parameters for the first halo model, will not update the model if they are the same as the pre-initialised model.

        :type: dict
        """
        return val

    @parameter('param')
    def _model_2_params(self, val):
        """
        Parameters for the second halo model.

        :type: dict
        """
        return val

    @parameter('param')
    def fraction_z(self, val):
        """
        Redshifts of the red/blue fraction of galaxies in sample.

        :type: array_like
        """
        return val

    @parameter('param')
    def fraction(self, val):
        """
        Red/blue fraction of galaxies in sample as a function of redshift.

        :type: array_like
        """
        return val

    @parameter('param')
    def z(self, val):
        """
        Output redshifts to which to interpolate and optionally extrapolate power spectra.

        :type: array_like
        """
        return val

    @parameter('param')
    def k(self, val):
        """
        Output k-vector to which to interpolate and optionally extrapolate power spectra.

        :type: array_like
        """
        return val

    @parameter('param')
    def extrapolate_option(self, val):
        """
        Extrapolation option to pass to interp1d.

        :type: str or float
        """
        return val

    @cached_property
    def frac_1(self):
        """
        Calculate the fraction for the first model.

        If no fraction is given, it assumes the first model is 100% and the second is 0%.
        This is useful for cases where only one model is used.

        Returns
        -------
        ndarray
            Fraction for the first model.
        """
        if self.fraction is None:
            return np.ones_like(self.z)
        f = interp1d(
            self.fraction_z,
            self.fraction,
            kind='linear',
            fill_value='extrapolate',
            bounds_error=False,
            axis=0,
        )
        return f(self.z)

    @cached_property
    def frac_2(self):
        """
        Calculate the fraction for the second model.

        Returns
        -------
        ndarray
            Fraction for the second model.
        """
        if self.fraction is None:
            return np.zeros_like(self.z)
        return 1.0 - self.frac_1

    @cached_property
    def power_1(self):
        """
        First Halo Model.

        Returns
        -------
        Spectra
            Instance of Spectra for the first halo model.
        """
        if self.model is None:
            return Spectra(**self._model_1_params)
        return self.model

    @cached_property
    def power_2(self):
        """
        Second Halo Model.

        We use the update method so that the second model does not have to recalculate
        all the methods if they are the same as the first one.

        Returns
        -------
        Spectra or None
            Instance of Spectra for the second halo model.
        """
        if self._model_2_params is None:
            spectra2 = None
        else:
            spectra2 = self.power_1.clone()
            spectra2.update(**self._model_2_params)
        return spectra2

    def results(self, requested_spectra, requested_components):
        """
        Calculate and store the results for the requested spectra and components.

        Parameters
        ----------
        requested_spectra : list
            List of spectra modes to calculate.
        requested_components : list
            List of components to calculate for each spectrum.

        Returns
        -------
        PowerSpectrumResults
            PowerSpectrumResults atributes attached to parent class
        """
        for mode in requested_spectra:
            collected_spectra = {}
            for component in requested_components:
                p1 = getattr(self.power_1, f'power_spectrum_{mode}')
                p1_component = getattr(p1, f'pk_{component}')
                extrapolated_p1 = self.extrapolate_spectra(
                    self.z,
                    self.k,
                    self.power_1.z_vec,
                    self.power_1.k_vec,
                    p1_component,
                    extrapolate_option=self.extrapolate_option,
                )
                if self.power_2 is not None:
                    p2 = getattr(self.power_2, f'power_spectrum_{mode}')
                    p2_component = getattr(p2, f'pk_{component}')
                    extrapolated_p2 = self.extrapolate_spectra(
                        self.z,
                        self.k,
                        self.power_2.z_vec,
                        self.power_2.k_vec,
                        p2_component,
                        extrapolate_option=self.extrapolate_option,
                    )
                else:
                    extrapolated_p2 = np.zeros_like(extrapolated_p1)
                added_power = self.add_spectra(extrapolated_p1, extrapolated_p2, mode)
                collected_spectra[f'pk_{component}'] = added_power
            # Create a PowerSpectrumResult object to hold the results
            power_spectrum_result = PowerSpectrumResult(**collected_spectra)
            setattr(self, f'power_spectrum_{mode}', power_spectrum_result)

    def extrapolate_spectra(self, z_ext, k_ext, z_in, k_in, power, extrapolate_option):
        """
        Extrapolate the power spectra to the desired output grid.

        Parameters
        ----------
        z_ext : array_like
            Output redshifts.
        k_ext : array_like
            Output k-vector.
        z_in : array_like
            Input redshifts.
        k_in : array_like
            Input k-vector.
        power : array_like
            Power spectra to extrapolate.
        extrapolate_option : str or float
            Extrapolation option to pass to interp1d.

        Returns
        -------
        ndarray
            Extrapolated power spectra.
        """
        inter_func_z = interp1d(
            z_in,
            power,
            kind='linear',
            fill_value=extrapolate_option,
            bounds_error=False,
            axis=1,
        )
        pk_tot_ext_z = inter_func_z(z_ext)

        inter_func_k = interp1d(
            np.log10(k_in),
            pk_tot_ext_z,
            kind='linear',
            fill_value='extrapolate',
            bounds_error=False,
            axis=2,
        )
        pk_tot_ext = inter_func_k(np.log10(k_ext))
        return pk_tot_ext

    def add_spectra(self, pk_1, pk_2, mode):
        """
        Add the power spectra of the two models.

        TODO: Add the cross terms
        Not valid / implemented for matter-intrinsic, matter-matter

        Parameters
        ----------
        pk_1 : array_like
            Power spectrum of the first model.
        pk_2 : array_like
            Power spectrum of the second model.
        mode : str
            Mode of the power spectrum.

        Returns
        -------
        ndarray
            Combined power spectrum.
        """
        if mode == 'mm':
            return pk_1
        if mode in ['gm', 'mi']:
            pk_tot = (
                self.frac_1[:, np.newaxis] * pk_1
                + (1.0 - self.frac_1[:, np.newaxis]) * pk_2
            )
        else:
            pk_tot = (
                self.frac_1[:, np.newaxis] ** 2.0 * pk_1
                + (1.0 - self.frac_1[:, np.newaxis]) ** 2.0 * pk_2
            )
        return pk_tot



import pytest
import numpy as np
from onepower import PowerSpectrumResult, Spectra


@pytest.fixture
def setup_data():
    rng = np.random.default_rng(seed=42)
    k_vec = np.logspace(-4, 4, 100)
    z_vec = np.linspace(0.0, 3.0, 15)
    mass = np.logspace(12, 15, 100)
    dndlnm = rng.random((15, 100))
    halo_bias = rng.random((15, 100))
    matter_power_lin = rng.random((15, 100))
    matter_power_nl = rng.random((15, 100))
    return k_vec, z_vec, mass, dndlnm, halo_bias, matter_power_lin, matter_power_nl


@pytest.fixture
def spectra(setup_data):
    k_vec, z_vec, mass, dndlnm, halo_bias, matter_power_lin, matter_power_nl = (
        setup_data
    )
    return Spectra(
        Mmin=12,
        Mmax=15,
        dlog10m=(15 - 12) / 100,
        matter_power_lin=matter_power_lin,
        matter_power_nl=matter_power_nl,
    )


def test_spectra_initialization_and_params_none(setup_data):
    _, _, _, _, _, matter_power_lin, matter_power_nl = setup_data
    spectra = Spectra(
        matter_power_lin=matter_power_lin, matter_power_nl=matter_power_nl
    )
    assert spectra.matter_power_lin.shape == matter_power_lin.shape
    assert spectra.matter_power_nl.shape == matter_power_nl.shape

    spectra_none = Spectra(hod_model=None, poisson_model=None)
    assert spectra_none.hod_model is None
    assert spectra_none.poisson_model is None


def test_spectra_properties_and_zvec_kvec_shape(setup_data, spectra):
    k_vec, z_vec, mass, _, _, matter_power_lin, matter_power_nl = setup_data

    assert spectra._beta_nl_array is None
    assert spectra._pk_lin.shape == (len(z_vec), len(k_vec))
    assert spectra._pk_nl.shape == (len(z_vec), len(k_vec))
    assert spectra.peff is None

    spectra_zvec_kvec = Spectra(
        k_vec=k_vec[:-2],
        z_vec=z_vec[:-2],
        matter_power_lin=matter_power_lin,
        matter_power_nl=matter_power_nl,
        fortuna=True,
        response=True,
    )
    with pytest.raises(ValueError):
        spectra_zvec_kvec._pk_lin
    with pytest.raises(ValueError):
        spectra_zvec_kvec._pk_nl


def test_spectra_bnl_related_properties(setup_data, spectra):
    k_vec, _, mass, _, _, _, _ = setup_data

    spectra.update(bnl=True)
    assert spectra.calc_bnl.shape == (1, len(mass), len(mass), len(k_vec))
    assert spectra.I12 is not None
    assert spectra.I21 is not None
    assert spectra.I22 is not None
    assert isinstance(spectra.power_spectrum_mm, PowerSpectrumResult)
    assert isinstance(spectra.power_spectrum_gg, PowerSpectrumResult)
    assert isinstance(spectra.power_spectrum_gm, PowerSpectrumResult)
    assert isinstance(spectra.power_spectrum_mi, PowerSpectrumResult)
    assert isinstance(spectra.power_spectrum_ii, PowerSpectrumResult)
    assert isinstance(spectra.power_spectrum_gi, PowerSpectrumResult)

    spectra.update(bnl=False)
    assert spectra.I12 is None
    assert spectra.I21 is None
    assert spectra.I22 is None


def test_spectra_hod_and_fstar_matter(setup_data, spectra):
    _, _, mass, _, _, _, _ = setup_data

    assert spectra.hod_mm is None
    assert spectra.fstar_mm.shape == (1, len(spectra.z_vec), len(mass))


def test_spectra_matter_profiles_and_truncation_properties(setup_data, spectra):
    _, z_vec, mass, _, _, _, _ = setup_data

    assert spectra.matter_profile.shape == (
        1,
        len(z_vec),
        len(spectra.k_vec),
        len(mass),
    )
    assert spectra.matter_profile_2h.shape == (
        1,
        len(z_vec),
        len(spectra.k_vec),
        len(mass),
    )

    k_vec, _, _, _, _, _, _ = setup_data
    assert spectra.one_halo_truncation.shape == (len(k_vec),)
    assert spectra.two_halo_truncation.shape == (len(k_vec),)
    assert spectra.one_halo_truncation_ia.shape == (len(k_vec),)
    assert spectra.two_halo_truncation_ia.shape == (len(k_vec),)

    spectra.update(
        one_halo_ktrunc=None,
        two_halo_ktrunc=None,
        one_halo_ktrunc_ia=None,
        two_halo_ktrunc_ia=None,
    )
    assert np.allclose(spectra.one_halo_truncation, np.ones_like(k_vec))
    assert np.allclose(spectra.two_halo_truncation, np.ones_like(k_vec))
    assert np.allclose(spectra.one_halo_truncation_ia, np.ones_like(k_vec))
    assert np.allclose(spectra.two_halo_truncation_ia, np.ones_like(k_vec))


def test_spectra_power_spectrum_properties_and_hod_observable_properties(
    setup_data, spectra
):
    _, z_vec, mass, _, _, _, _ = setup_data

    assert isinstance(spectra.power_spectrum_lin, PowerSpectrumResult)
    assert isinstance(spectra.power_spectrum_mm, PowerSpectrumResult)

    assert spectra.hod is not None
    assert spectra.fstar.shape == (1, len(z_vec), len(mass))
    assert spectra.mass_avg.shape == (1, len(z_vec))

    spectra.update(compute_observable=True)
    assert spectra.obs is not None
    assert spectra.obs_func is not None
    spectra.update(compute_observable=False)


def test_spectra_galaxy_profiles_and_terms_and_power_spectra(setup_data, spectra):
    _, z_vec, mass, _, _, _, _ = setup_data

    assert spectra.central_galaxy_profile.shape == (
        1,
        len(z_vec),
        len(spectra.k_vec),
        len(mass),
    )
    assert spectra.satellite_galaxy_profile.shape == (
        1,
        len(z_vec),
        len(spectra.k_vec),
        len(mass),
    )

    assert spectra.Ic_term.shape == (1, len(z_vec), len(spectra.k_vec))
    assert spectra.Is_term.shape == (1, len(z_vec), len(spectra.k_vec))
    assert isinstance(spectra.power_spectrum_gg, PowerSpectrumResult)
    assert isinstance(spectra.power_spectrum_gm, PowerSpectrumResult)


def test_spectra_alignment_properties_and_profiles_and_terms(setup_data, spectra):
    _, z_vec, mass, _, _, _, _ = setup_data

    assert spectra.alignment_class is not None
    assert spectra.beta_cen is not None
    assert spectra.beta_sat is not None
    assert spectra.mpivot_cen is not None
    assert spectra.mpivot_sat is not None
    assert spectra.alignment_gi is not None
    assert spectra.alignment_amplitude_2h.shape == (len(z_vec), 1)
    assert spectra.alignment_amplitude_2h_II.shape == (len(z_vec), 1)
    assert spectra.C1.shape == (len(z_vec), 1, 1)

    assert spectra.wkm_sat.shape == (1, len(z_vec), len(mass), len(spectra.k_vec))
    assert spectra.central_alignment_profile.shape == (1, len(z_vec), 1, len(mass))
    assert spectra.satellite_alignment_profile.shape == (
        1,
        len(z_vec),
        len(spectra.k_vec),
        len(mass),
    )
    assert spectra.Ic_align_term.shape == (1, len(z_vec), 1)
    assert spectra.Is_align_term.shape == (1, len(z_vec), len(spectra.k_vec))


def test_spectra_alignment_power_spectra_and_mead(setup_data):
    _, _, mass, _, _, _, _ = setup_data

    spectra_mead = Spectra(mead_correction='nofeedback')
    assert isinstance(spectra_mead.power_spectrum_mm, PowerSpectrumResult)
    assert isinstance(spectra_mead.power_spectrum_gg, PowerSpectrumResult)
    assert isinstance(spectra_mead.power_spectrum_gm, PowerSpectrumResult)
    assert isinstance(spectra_mead.power_spectrum_mi, PowerSpectrumResult)
    assert isinstance(spectra_mead.power_spectrum_ii, PowerSpectrumResult)
    assert isinstance(spectra_mead.power_spectrum_gi, PowerSpectrumResult)

    spectra_mead = Spectra(mead_correction='feedback')
    assert isinstance(spectra_mead.power_spectrum_mm, PowerSpectrumResult)
    assert isinstance(spectra_mead.power_spectrum_gg, PowerSpectrumResult)
    assert isinstance(spectra_mead.power_spectrum_gm, PowerSpectrumResult)
    assert isinstance(spectra_mead.power_spectrum_mi, PowerSpectrumResult)
    assert isinstance(spectra_mead.power_spectrum_ii, PowerSpectrumResult)
    assert isinstance(spectra_mead.power_spectrum_gi, PowerSpectrumResult)

    spectra_mead = Spectra(
        mead_correction='fit', Mmin=12, Mmax=15, dlog10m=(15 - 12) / 100
    )
    assert spectra_mead.hod_mm is not None
    assert spectra_mead.fstar_mm.shape == (1, len(spectra_mead.z_vec), len(mass))
    assert isinstance(spectra_mead.power_spectrum_mm, PowerSpectrumResult)
    assert isinstance(spectra_mead.power_spectrum_gm, PowerSpectrumResult)
    assert isinstance(spectra_mead.power_spectrum_mi, PowerSpectrumResult)
    assert isinstance(spectra_mead.power_spectrum_ii, PowerSpectrumResult)
    assert isinstance(spectra_mead.power_spectrum_gi, PowerSpectrumResult)


def test_fortuna_and_response(setup_data):
    k_vec, z_vec, mass, _, _, matter_power_lin, matter_power_nl = setup_data
    spectra = Spectra(
        k_vec=k_vec,
        z_vec=z_vec,
        matter_power_lin=matter_power_lin,
        matter_power_nl=matter_power_nl,
        fortuna=True,
        response=True,
    )

    assert spectra._beta_nl_array is None
    assert spectra._pk_lin.shape == (len(z_vec), len(k_vec))
    assert spectra._pk_nl.shape == (len(z_vec), len(k_vec))
    assert spectra.peff is not None
    assert isinstance(spectra.power_spectrum_mm, PowerSpectrumResult)
    assert isinstance(spectra.power_spectrum_gg, PowerSpectrumResult)
    assert isinstance(spectra.power_spectrum_gm, PowerSpectrumResult)
    assert isinstance(spectra.power_spectrum_mi, PowerSpectrumResult)
    assert isinstance(spectra.power_spectrum_ii, PowerSpectrumResult)
    assert isinstance(spectra.power_spectrum_gi, PowerSpectrumResult)
